"""
Functions written in pure python
"""
from copy import deepcopy, copy
import warnings

import narwhals.stable.v2 as nw

# Functions to deal with value labels

def set_value_labels(dataframe, metadata, formats_as_category=True, formats_as_ordered_category=False):
    """
    Changes the values in the dataframe according to the value formats in the metadata.
    It will return a copy of the dataframe. If no appropiate formats were found, the result will be an unchanged copy
    of the original dataframe.

    Parameters
    ----------
        dataframe : pandas dataframe
            resulting from parsing a file
        metadata : dictionary
            resulting from parsing a file
        formats_as_category : bool, optional
            defaults to True. If True the variables having formats will be transformed into pandas categories.
        formats_as_ordered_category : bool, optional
            defaults to False. If True the variables having formats will be transformed into pandas ordered categories.
            it has precedence over formats_as_category, meaning if this is True, it will take effect irrespective of
            the value of formats_as_category.

    Returns
    -------
        df_copy : pandas dataframe
            a copy of the original dataframe with the values changed, if appropiate formats were found, unaltered
            otherwise
    """

    df_copy = nw.from_native(dataframe).clone()

    if metadata.value_labels and metadata.variable_to_label:
        for var_name, label_name in metadata.variable_to_label.items():
            labels = metadata.value_labels.get(label_name)
            if labels:
                labels = deepcopy(labels)
                if var_name in df_copy.columns:
                    # unique does not work for polars Object
                    if not df_copy.implementation.is_pandas() and df_copy[var_name].dtype == nw.Object:
                        unvals = list(set(df_copy[var_name].to_list()))
                    else:
                        unvals = df_copy[var_name].unique()
                    for uval in unvals:
                        if uval not in labels:
                            labels[uval] = uval
                    # if all values are null, there will be nothing to replace. However we cannot do replace_strict on null dtype, it raises an error
                    if not df_copy.implementation.is_pandas() and (len(df_copy[var_name])==df_copy[var_name].null_count()):
                        continue
                    # replace_strict requires that all the values are in the map. Could not get map_batches or when/then/otherwise to work
                    elif not df_copy.implementation.is_pandas() and (df_copy[var_name].dtype==nw.Object or not all([type(v)==type(list(labels.values())[0]) for v in labels.values() if v is not None])):
                        # polars is very difficult to convince to mix strings and numbers, so we have to do it this way
                        temp = [labels[x] for x in df_copy[var_name]]
                        newser = nw.new_series(name=var_name, values= temp, dtype=nw.Object, backend=df_copy.implementation) 
                        df_copy = df_copy.with_columns(newser.alias(var_name))
                        if formats_as_category or formats_as_ordered_category:
                            msg = f"You requested formats_as_category=True or formats_as_ordered_category=True, but it was not possible to cast variable '{var_name}' to category"
                            warnings.warn(msg, RuntimeWarning)
                            continue
                    # not sure if we get into this situation ever or what would exactly happen, maybe this is not needed?
                    elif not df_copy.implementation.is_pandas() and df_copy[var_name].dtype==nw.Unknown:
                        msg = f"It was not possible to apply value formats to variable '{var_name}' due to unknown/not supported data type"
                        warnings.warn(msg, RuntimeWarning)
                        continue
                    else:
                        df_copy = df_copy.with_columns(nw.col(var_name).replace_strict(labels))
                    if formats_as_ordered_category:
                        categories = list(set(labels.values()))
                        original_values = list(labels.keys())
                        original_values.sort()
                        revdict= dict()
                        for orival in original_values:
                            curcat = labels.get(orival)
                            if not revdict.get(curcat):
                                revdict[curcat] = orival
                        categories.sort(key=revdict.get)
                        df_copy = df_copy.with_columns(nw.col(var_name).cast(nw.Enum(categories)))
                    elif formats_as_category:
                        df_copy = df_copy.with_columns(nw.col(var_name).cast(nw.Categorical))


    return df_copy.to_native()

def set_catalog_to_sas(sas_dataframe, sas_metadata, catalog_metadata, formats_as_category=True,  
                       formats_as_ordered_category=False):
    """
    Changes the values in the dataframe and sas_metadata according to the formats in the catalog.
    It will return a copy of the dataframe and metadata. If no appropriate formats were found, the result will
    be an unchanged copy of the original dataframe.

    Parameters
    ----------
        sas_dataframe : pandas dataframe
            resulting from parsing a sas7bdat file
        sas_metadata : pyreadstat metadata object
            resulting from parsing a sas7bdat file
        catalog_metadata : pyreadstat metadata object
            resulting from parsing a sas7bcat (catalog) file
        formats_as_category : bool, optional
            defaults to True. If True the variables having formats will be transformed into pandas categories.
        formats_as_ordered_category : bool, optional
            defaults to False. If True the variables having formats will be transformed into pandas ordered categories.
            it has precedence over formats_as_category, meaning if this is True, it will take effect irrespective of
            the value of formats_as_category.

    Returns
    -------
        df_copy : pandas dataframe
            a copy of the original dataframe with the values changed, if appropriate formats were found, unaltered
            otherwise
        metadata : dict
            a copy of the original sas_metadata enriched with catalog information if found, otherwise unaltered
    """

    if catalog_metadata.value_labels and sas_metadata.variable_to_label:
        catalog_metadata_copy = deepcopy(catalog_metadata)
        metadata = deepcopy(sas_metadata)
        metadata.value_labels = catalog_metadata_copy.value_labels
        df_copy = set_value_labels(sas_dataframe, metadata, formats_as_category=formats_as_category,
                                   formats_as_ordered_category=formats_as_ordered_category)

        variable_value_labels = dict()
        for var_name, var_label in metadata.variable_to_label.items():
            current_labels = catalog_metadata_copy.value_labels.get(var_label)
            if current_labels:
                variable_value_labels[var_name] = current_labels
        metadata.variable_value_labels = variable_value_labels

    else:
        #df_copy = sas_dataframe.copy()
        df_copy = nw.from_native(sas_dataframe).clone().to_native()
        metadata = deepcopy(sas_metadata)

    return df_copy, metadata




#-*- coding: utf-8 -*-
# #############################################################################
# Copyright 2018 Hoffmann-La Roche
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# #############################################################################

from datetime import datetime, timedelta, date
import unittest
import os
import sys
import shutil
import multiprocessing as mp
import tempfile
import zipfile
import io

import pandas as pd
import narwhals as nw
import numpy as np
import polars as pl

is_pathlib_available = False
try:
    from pathlib import Path
    is_pathlib_available = True
except:
    pass

is_pandas_version_newer_than_2 = int(pd.__version__.split(".")[0])>2

class TestBasic(unittest.TestCase):
    """
    Test suite for pyreadstat.
    """

    def _prepare_data(self):
        
        #backend = 'polars'
        global backend
        self.backend = backend

        self.script_folder = os.path.dirname(os.path.realpath(__file__))
        self.parent_folder = os.path.split(self.script_folder)[0]
        self.data_folder = os.path.join(self.parent_folder, "test_data")
        self.basic_data_folder = os.path.join(self.data_folder, "basic")
        self.catalog_data_folder = os.path.join(self.data_folder, "sas_catalog")
        self.international_data_folder = os.path.join(self.data_folder, "ínternátionál")
        self.missing_data_folder = os.path.join(self.data_folder, "missing_data")
        self.mr_data_folder = os.path.join(self.data_folder, "multiple_response")
        self.write_folder = os.path.join(self.data_folder, "write")
        if not os.path.isdir(self.write_folder):
            os.makedirs(self.write_folder)


        # basic
        pandas_csv = os.path.join(self.basic_data_folder, "sample.csv")
        #df_pandas = pl.read_csv(pandas_csv, try_parse_dates=True)
        kwds = {}
        if backend == "polars":
            kwds["try_parse_dates"] = True
        df_pandas = nw.read_csv(pandas_csv, backend=backend, **kwds)
        if backend == "pandas":
            df_pandas = df_pandas.to_native()
            df_pandas["mydate"] = [datetime.strptime(x, '%Y-%m-%d').date() if type(x) == str else float('nan') for x in df_pandas["mydate"]]
            df_pandas["dtime"] = [datetime.strptime(x, '%Y-%m-%dT%H:%M:%S.000000') if type(x) == str else float('nan') for x in df_pandas["dtime"]]
            df_pandas["mytime"] = [datetime.strptime(x, '%H:%M:%S.000000').time() if type(x) == str else float('nan') for x in df_pandas["mytime"]]
            df_pandas = nw.from_native(df_pandas)
        df_pandas = df_pandas.with_columns(nw.col("myord", "mylabl").cast(nw.Float64))
        self.df_pandas = df_pandas.to_native()
        # skip some columns
        self.usecols = ['mynum', 'myord']
        cols_to_drop = list(set(df_pandas.columns) - set(self.usecols))
        self.df_usecols = df_pandas.drop(cols_to_drop).to_native()#, axis=1)

        # no dates
        nodates_spss_csv = os.path.join(self.basic_data_folder, "sample_nodate_spss.csv")
        df_nodates_spss = nw.read_csv(nodates_spss_csv, backend=backend)
        df_nodates_spss = df_nodates_spss.with_columns(nw.col("myord", "mylabl").cast(nw.Float64))
        self.df_nodates_spss = df_nodates_spss.to_native()

        nodates_sastata_csv = os.path.join(self.basic_data_folder, "sample_nodate_sas_stata.csv")
        df_nodates_sastata = nw.read_csv(nodates_sastata_csv, backend=backend)
        df_nodates_sastata = df_nodates_sastata.with_columns(nw.col("myord", "mylabl").cast(nw.Float64))
        self.df_nodates_sastata = df_nodates_sastata.to_native()

        # xport files v5 vs v8
        self.xptv5v8 = nw.from_dict({'i': [float(x) for x in range(1,11)]}, backend=backend).to_native()
        # formatted
        mylabl_format = {1.0:"Male", 2.0:"Female"}
        myord_format = {1.0:"low", 2.0:"medium", 3.0:"high"}
        df_pandas_formatted = df_pandas.clone()
        df_pandas_formatted = df_pandas_formatted.with_columns(nw.col("mylabl").replace_strict(mylabl_format))
        df_pandas_formatted = df_pandas_formatted.with_columns(nw.col("myord").replace_strict(myord_format))
        df_pandas_formatted = df_pandas_formatted.with_columns(nw.col("myord", "mylabl").cast(nw.Categorical))
        self.df_pandas_formatted = df_pandas_formatted.to_native()
        # sas formatted
        sas_formatted = os.path.join(self.catalog_data_folder, "sas_formatted.csv")
        df_sas = nw.read_csv(sas_formatted, backend=backend)
        df_sas = df_sas.with_columns(nw.col("SEXA", "SEXB").cast(nw.Categorical))
        self.df_sas_format = df_sas.to_native()

        # missing data
        pandas_missing_sav_csv = os.path.join(self.basic_data_folder, "sample_missing.csv")
        kwds = {}
        if backend == "pandas":
            kwds["na_values"] = "#NULL!"
            kwds["keep_default_na"] = False
        if backend == "polars":
            kwds["null_values"] = "#NULL!"
            kwds["try_parse_dates"] = True
            kwds["missing_utf8_is_empty_string"] = True
        df_missing_sav = nw.read_csv(pandas_missing_sav_csv, backend=backend, **kwds )
        df_missing_sav = df_missing_sav.to_native()
        if backend == "pandas":
            df_missing_sav["mydate"] = [datetime.strptime(x, '%Y-%m-%d').date() if type(x) == str else float('nan') for x in
                                    df_missing_sav["mydate"]]
            df_missing_sav["dtime"] = [datetime.strptime(x, '%Y-%m-%dT%H:%M:%S.000000') if type(x) == str else float('nan') for x
                              in df_missing_sav["dtime"]]
            df_missing_sav["mytime"] = [datetime.strptime(x, '%H:%M:%S.000000').time() if type(x) == str else float('nan') for x
                               in df_missing_sav["mytime"]]
        self.df_missing_sav = df_missing_sav
        # user missing
        pandas_missing_user_sav_csv = os.path.join(self.basic_data_folder, "sample_missing_user.csv")
        df_user_missing_sav = nw.read_csv(pandas_missing_user_sav_csv, backend=backend, **kwds)
        df_user_missing_sav = df_user_missing_sav.to_native()
        if backend == "pandas":
            df_user_missing_sav["mydate"] = [datetime.strptime(x, '%Y-%m-%d').date() if type(x) == str else float('nan') for x in
                                             df_user_missing_sav["mydate"]]
            df_user_missing_sav["dtime"] = [datetime.strptime(x, '%Y-%m-%dT%H:%M:%S.000000') if type(x) == str else float('nan')
                                       for x in df_user_missing_sav["dtime"]]
            df_user_missing_sav["mytime"] = [datetime.strptime(x, '%H:%M:%S.000000').time() if type(x) == str else float('nan')
                                        for x in df_user_missing_sav["mytime"]]
        self.df_user_missing_sav = df_user_missing_sav

        # dates
        sas_dates = os.path.join(self.basic_data_folder, "dates.csv")
        kwds = {}
        if backend == "polars":
            kwds["try_parse_dates"] = True
        df_dates_raw = nw.read_csv(sas_dates,backend=backend, **kwds)
        df_dates1 = df_dates_raw.clone()
        if backend == "pandas":
            df_dates1 = df_dates1.to_native()
            df_dates1["date"] = pd.to_datetime(df_dates1["date"])
            df_dates1["dtime"] = pd.to_datetime(df_dates1["dtime"])
            df_dates1["time"] = pd.to_datetime(df_dates1["time"], format='%H:%M:%S')
            df_dates1["time"] = df_dates1["time"].apply(lambda x: x.time())
        else:
            df_dates1 = df_dates1.with_columns(nw.col("date").cast(nw.Date)).to_native()

        self.df_sas_dates_as_pandas = df_dates1

        if backend == "pandas":
            df_dates2 = df_dates1.copy()
            df_dates2["date"] = df_dates2["date"].apply(lambda x: x.date())
            self.df_sas_dates = df_dates2
            self.df_sas_dates2 = pd.concat([self.df_sas_dates, pd.DataFrame([[np.nan, pd.NaT, np.nan]],columns=["date", "dtime", "time"])], ignore_index=True)
        else:
            df_dates2 = df_dates_raw.clone()#.with_columns(nw.col("time").cast(nw.Time))
            self.df_sas_dates = df_dates2.to_native()
            #schema = {"date": nw.Date, "dtime": nw.Datetime("ns"), "time": nw.Time()}
            self.df_sas_dates2 = nw.concat([df_dates2, nw.from_dict({"date":[None], "dtime":[None], "time":[None]}, backend=backend)]).to_native() #, schema=schema
        # character column with nan and object column with nan (object pyreadstat writer doesn't know what to do with)
        if backend == "pandas":
            self.df_charnan = pd.DataFrame([[0,np.nan,np.nan],[1,"test", timedelta]], columns = ["integer", "string", "object"])
        else:
            self.df_charnan = nw.from_dict({'integer': [0, 1], "string": [None, "test"], "object": [None, timedelta]}, backend=backend).to_native()

        # long string

        self.df_longstr = nw.from_dict({
                "v1": [
                     """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Maecenas ac pretium sem. Fusce aliquet
                    augue rhoncus consequat pulvinar. In est ex, porta congue diam sed, laoreet suscipit purus. Phasellus mollis
                    lobortis tellus at vehicula. Etiam egestas augue id massa bibendum volutpat id et ipsum. Praesent ut lorem
                    rhoncus, pharetra risus sed, pharetra sem. In pulvinar egestas erat, id condimentum tortor tempus sed. Duis
                    ornare lacus ut ligula congue, non convallis urna dignissim. Etiam vehicula turpis sit amet nisi finibus
                    laoreet. Duis molestie consequat nulla, non lobortis est tempus sit amet. Quisque elit est,
                    congue non commodo vitae, porttitor ac erat. """ *3,
                     "fgsdghshsgh",
                     "gsfdgsdg",
                ],
                "v2": [
                     "gsfdgsfdgsfg",
                     "fgsdghshsgh",
                     "gsfdgsdg",
                ],

            }, backend=backend).to_native()

    def setUp(self):

        # set paths
        self._prepare_data()

    def test_sas7bdat(self):
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample.sas7bdat"), output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas))
        #self.assertTrue(meta.creation_time==datetime(2018, 8, 16, 18, 21, 52))
        #self.assertTrue(meta.modification_time==datetime(2018, 8, 16, 18, 21, 52))

    def test_sas7bdat_bincompressed(self):
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample_bincompressed.sas7bdat"), output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas))

    def test_sas7bdat_metaonly(self):
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample.sas7bdat"), output_format=self.backend)
        df2, meta2 = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample.sas7bdat"), metadataonly=True, output_format=self.backend)
        self.assertTrue(len(df2)==0)
        self.assertTrue(meta.number_columns == meta2.number_columns)
        self.assertTrue(meta.number_rows == meta2.number_rows)
        self.assertTrue(meta.column_names == meta2.column_names)
        self.assertTrue(meta.column_labels == meta2.column_labels)
        self.assertTrue(meta.readstat_variable_types["mychar"]=="string")
        self.assertTrue(meta.readstat_variable_types["myord"]=="double")
        self.assertTrue(meta.readstat_variable_types["dtime"]=="double")

    def test_sas7bdat_usecols(self):
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample.sas7bdat"), usecols=self.usecols, output_format=self.backend)
        self.assertTrue(df.equals(self.df_usecols))
        self.assertTrue(meta.number_columns == len(self.usecols))
        self.assertTrue(meta.column_names == self.usecols)
        
    def test_sas7bdat_international(self):
        """
        On windows, paths with international characters are problematic. This is verifying that it is working as expected
        """
        # in addition, this works only in python 3
        if sys.version_info[0]>2:
            df, meta = pyreadstat.read_sas7bdat(os.path.join(self.international_data_folder, "sample.sas7bdat"), output_format=self.backend)
            self.assertTrue(df.equals(self.df_pandas))
            self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
            self.assertTrue(meta.number_rows == len(self.df_pandas))

    def test_sas7bdat_nodates(self):
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample.sas7bdat"), disable_datetime_conversion=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_nodates_sastata))

    def test_sas7bdat_chunk(self):
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "sample.sas7bdat"), row_limit = 2, row_offset =1, output_format=self.backend)
        df_pandas = nw.from_native(self.df_pandas)[1:3].to_native()
        if self.backend == "pandas":
            df_pandas = df_pandas.reset_index(drop=True)
            df_pandas["dtime"] = pd.to_datetime(df_pandas["dtime"])
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))

    def test_xport(self):
        df, meta = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sample.xpt"), output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(self.df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas))
        self.assertTrue(meta.number_rows==len(df))
        #self.assertTrue(meta.creation_time==datetime(2018, 8, 14, 10, 55, 46))
        #self.assertTrue(meta.modification_time==datetime(2018, 8, 14, 10, 55, 46))

    def test_xport_v5(self):
        df, meta = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sas.xpt5"), output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(self.xptv5v8))
        self.assertTrue(meta.number_rows==len(df))

    def test_xport_v8(self):
        df, meta = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sas.xpt8"), output_format=self.backend)
        self.assertTrue(df.equals(self.xptv5v8))
        self.assertTrue(meta.number_rows==len(df))

    def test_xport_metaonly(self):
        df, meta = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sample.xpt"), output_format=self.backend)
        df2, meta2 = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sample.xpt"), metadataonly=True, output_format=self.backend)
        self.assertTrue(len(df2)==0)
        self.assertTrue(meta.number_columns == meta2.number_columns)
        self.assertTrue(meta2.number_rows is None)
        self.assertTrue(meta.column_names == meta2.column_names)
        self.assertTrue(meta.column_labels == meta2.column_labels)

    def test_xport_usecols(self):
        # Currently readstat does not support skipping cols for XPT files,
        usecols = [x.upper() for x in self.usecols]
        df, meta = pyreadstat.pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sample.xpt"), usecols=usecols, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(self.df_usecols))
        self.assertTrue(meta.number_columns == len(self.usecols))

    def test_xport_nodates(self):
        df, meta = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sample.xpt"), disable_datetime_conversion=True, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(self.df_nodates_sastata))

    def test_xport_chunks(self):
        df, meta = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "sample.xpt"), row_limit = 2, row_offset =1, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        df_pandas = nw.from_native(self.df_pandas)[1:3].to_native()
        if self.backend == "pandas":
            df_pandas = df_pandas.reset_index(drop=True)
            df_pandas["dtime"] = pd.to_datetime(df_pandas["dtime"])
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))

    def test_dta(self):
        # discard dtime and arrange time
        df, meta = pyreadstat.read_dta(os.path.join(self.basic_data_folder, "sample.dta"), output_format=self.backend)
        df_pandas = nw.from_native(self.df_pandas).clone()
        df_pandas = df_pandas.with_columns(nw.col("myord", "mylabl").cast(nw.Int64))
        df_pandas = df_pandas.to_native()
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))
        self.assertTrue(meta.readstat_variable_types["mychar"]=="string")
        self.assertTrue(meta.readstat_variable_types["myord"]=="int8")
        self.assertTrue(meta.readstat_variable_types["dtime"]=="double")
        #self.assertTrue(meta.creation_time==datetime(2018, 12, 17, 14, 53))
        #self.assertTrue(meta.modification_time==datetime(2018, 12, 17, 14, 53))

    def test_dta_metaonly(self):
        df, meta = pyreadstat.read_dta(os.path.join(self.basic_data_folder, "sample.dta"), output_format=self.backend)
        df2, meta2 = pyreadstat.read_dta(os.path.join(self.basic_data_folder, "sample.dta"), metadataonly=True, output_format=self.backend)
        self.assertTrue(len(df2)==0)
        self.assertTrue(meta.number_columns == meta2.number_columns)
        self.assertTrue(meta.number_rows == meta2.number_rows)
        self.assertTrue(meta.column_names == meta2.column_names)
        self.assertTrue(meta.column_labels == meta2.column_labels)

    def test_dta_usecols(self):
        df, meta = pyreadstat.read_dta(os.path.join(self.basic_data_folder, "sample.dta"), usecols=self.usecols, output_format=self.backend)
        #df_pandas = self.df_usecols.copy()
        df_pandas = nw.from_native(self.df_usecols).clone()
        #df_pandas["myord"] = df_pandas["myord"].astype(np.int64)
        df_pandas = df_pandas.with_columns(nw.col("myord").cast(nw.Int64))
        df_pandas = df_pandas.to_native()
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(self.usecols))
        self.assertTrue(meta.column_names == self.usecols)

    def test_dta_nodates(self):
        df, meta = pyreadstat.read_dta(os.path.join(self.basic_data_folder,"sample.dta"), disable_datetime_conversion=True, output_format=self.backend)
        #df_pandas = self.df_nodates_sastata
        df_pandas = nw.from_native(self.df_nodates_sastata)
        df_pandas = df_pandas.with_columns(nw.col("myord", "mylabl").cast(nw.Int64))
        df_pandas = df_pandas.with_columns(nw.col("dtime", "mytime")*1000)
        df_pandas = df_pandas.to_native()
        self.assertTrue(df.equals(df_pandas))

    def test_dta_chunks(self):
        # discard dtime and arrange time
        df, meta = pyreadstat.read_dta(os.path.join(self.basic_data_folder, "sample.dta"), row_limit = 2, row_offset =1, output_format = self.backend)
        df_pandas = nw.from_native(self.df_pandas)[1:3]
        df_pandas = df_pandas.with_columns(nw.col("myord", "mylabl").cast(nw.Int64))
        df_pandas = df_pandas.to_native()
        if self.backend == "pandas":
            df_pandas = df_pandas.reset_index(drop=True)
            df_pandas["dtime"] = pd.to_datetime(df_pandas["dtime"])
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))

    def test_sav(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas))
        self.assertTrue(len(meta.notes)>0)
        self.assertTrue(meta.variable_display_width["mychar"]==9)
        self.assertTrue(meta.variable_storage_width["mychar"] == 8)
        self.assertTrue(meta.variable_measure["mychar"]=="nominal")
        self.assertTrue(meta.readstat_variable_types["mychar"]=="string")
        self.assertTrue(meta.readstat_variable_types["myord"]=="double")
        #self.assertTrue(meta.creation_time==datetime(2018, 8, 16, 17, 22, 33))
        #self.assertTrue(meta.modification_time==datetime(2018, 8, 16, 17, 22, 33))

    def test_sav_metaonly(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), output_format=self.backend)
        df2, meta2 = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), metadataonly=True, output_format=self.backend)
        self.assertTrue(len(df2)==0)
        self.assertTrue(meta.number_columns == meta2.number_columns)
        self.assertTrue(meta.number_rows == meta2.number_rows)
        self.assertTrue(meta.column_names == meta2.column_names)
        self.assertTrue(meta.column_labels == meta2.column_labels)
        self.assertTrue(len(meta2.notes) > 0)

    def test_sav_formatted(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), apply_value_formats=True, formats_as_category=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas_formatted))
        self.assertTrue(meta.number_columns == len(self.df_pandas_formatted.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas_formatted))
        self.assertTrue(len(meta.notes) > 0)

    def test_sav_usecols(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), usecols=self.usecols, output_format=self.backend)
        self.assertTrue(df.equals(self.df_usecols))
        self.assertTrue(meta.number_columns == len(self.usecols))
        self.assertTrue(meta.column_names == self.usecols)

    def test_sav_missing(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample_missing.sav"), output_format=self.backend)
        self.assertTrue(df.equals(self.df_missing_sav))
        self.assertTrue(meta.missing_ranges == {})
        df_user, meta_user = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample_missing.sav"), user_missing=True, output_format=self.backend)
        self.assertTrue(df_user.equals(self.df_user_missing_sav))
        self.assertTrue(meta_user.missing_ranges['mynum'][0]=={'lo': 2000.0, 'hi': 3000.0})
        # user missing with usecols
        df_user, meta_user = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample_missing.sav"), user_missing=True, usecols=["mynum", "mylabl"], output_format=self.backend)
        df_sub = self.df_user_missing_sav[["mynum", "mylabl"]]
        self.assertTrue(df_user.equals(df_sub))

    def test_sav_nodates(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), disable_datetime_conversion=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_nodates_spss))

    def test_sav_chunks(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), row_limit = 2, row_offset =1, output_format=self.backend)
        df_pandas = nw.from_native(self.df_pandas)[1:3].to_native()
        if self.backend == "pandas":
            #df_pandas = self.df_pandas.iloc[1:3,:].reset_index(drop=True)
            df_pandas = df_pandas.reset_index(drop=True)
            df_pandas["dtime"] = pd.to_datetime(df_pandas["dtime"])
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))
        self.assertTrue(len(meta.notes)>0)
        self.assertTrue(meta.variable_display_width["mychar"]==9)
        self.assertTrue(meta.variable_storage_width["mychar"] == 8)
        self.assertTrue(meta.variable_measure["mychar"]=="nominal")
    
    def test_sav_expand(self):
        src = os.path.join(self.basic_data_folder, "sample.sav")
        dst = "~/sample.sav"
        shutil.copyfile(src, os.path.expanduser(dst))
        df, meta = pyreadstat.read_sav(dst, output_format=self.backend)
        os.remove(os.path.expanduser(dst))
        self.assertTrue(df.equals(self.df_pandas))
       
    def test_zsav(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.zsav"), output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas))
        self.assertTrue(len(meta.notes) > 0)

    def test_zsav_metaonly(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.zsav"), output_format=self.backend)
        df2, meta2 = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), metadataonly=True, output_format=self.backend)
        self.assertTrue(len(df2)==0)
        self.assertTrue(meta.number_columns == meta2.number_columns)
        self.assertTrue(meta.number_rows == meta2.number_rows)
        self.assertTrue(meta.column_names == meta2.column_names)
        self.assertTrue(meta.column_labels == meta2.column_labels)
        self.assertTrue(len(meta2.notes) > 0)

    def test_zsav_formatted(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.zsav"), apply_value_formats=True, formats_as_category=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas_formatted))
        self.assertTrue(meta.number_columns == len(self.df_pandas_formatted.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas_formatted))
        self.assertTrue(len(meta.notes) > 0)

    def test_zsav_usecols(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.zsav"), usecols=self.usecols, output_format=self.backend)
        self.assertTrue(df.equals(self.df_usecols))
        self.assertTrue(meta.number_columns == len(self.usecols))
        self.assertTrue(meta.column_names == self.usecols)

    def test_zsav_nodates(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.zsav"), disable_datetime_conversion=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_nodates_spss))

    def test_zsav_chunks(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.zsav"), row_limit = 2, row_offset =1, output_format=self.backend)
        df_pandas = nw.from_native(self.df_pandas)[1:3].to_native()
        if self.backend == "pandas":
            df_pandas = df_pandas.reset_index(drop=True)
            df_pandas["dtime"] = pd.to_datetime(df_pandas["dtime"])
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))
        self.assertTrue(len(meta.notes)>0)
        self.assertTrue(meta.variable_display_width["mychar"]==9)
        self.assertTrue(meta.variable_storage_width["mychar"] == 8)
        self.assertTrue(meta.variable_measure["mychar"]=="nominal")

    def test_por(self):
        df, meta = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"), output_format=self.backend)
        df_pandas_por = self.df_pandas
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(df_pandas_por))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas_por))
        self.assertTrue(len(meta.notes) > 0)
        #self.assertTrue(meta.creation_time==datetime(2018, 12, 16, 17, 28, 21))
        #self.assertTrue(meta.modification_time==datetime(2018, 12, 16, 17, 28, 21))

    def test_por_formatted(self):
        df, meta = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"), apply_value_formats=True, formats_as_category=True, output_format=self.backend)
        df_pandas_por = self.df_pandas_formatted
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(df_pandas_por))
        self.assertTrue(meta.number_columns == len(self.df_pandas_formatted.columns))
        self.assertTrue(meta.number_rows == len(df_pandas_por))
        self.assertTrue(len(meta.notes) > 0)

    def test_por_metaonly(self):
        df, meta = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"), output_format=self.backend)
        df2, meta2 = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"), metadataonly=True, output_format=self.backend)
        self.assertTrue(len(df2)==0)
        self.assertTrue(meta.number_columns == meta2.number_columns)
        self.assertTrue(meta2.number_rows is None)
        self.assertTrue(meta.column_names == meta2.column_names)
        self.assertTrue(meta.column_labels == meta2.column_labels)
        self.assertTrue(len(meta2.notes) > 0)

    def test_por_usecols(self):
        df, meta = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"),  usecols=["MYNUM"], output_format=self.backend)
        df_pandas_por = nw.from_native(self.df_pandas).select("mynum").to_native()
        #df_pandas_por = self.df_pandas[["mynum"]]
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(df_pandas_por))
        self.assertTrue(meta.number_columns == len(df_pandas_por.columns))

    def test_por_nodates(self):
        df, meta = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"), disable_datetime_conversion=True, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(self.df_nodates_spss))

    def test_por_chunks(self):
        df, meta = pyreadstat.read_por(os.path.join(self.basic_data_folder, "sample.por"), row_limit = 2, row_offset =1, output_format=self.backend)
        df_pandas_por = nw.from_native(self.df_pandas)[1:3].to_native()
        if self.backend == "pandas":
            df_pandas_por = df_pandas_por.reset_index(drop=True)
            df_pandas_por['dtime'] = pd.to_datetime(df_pandas_por.dtime)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(df_pandas_por))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas_por))
        self.assertTrue(len(meta.notes) > 0)

    def test_sas_catalog_win(self):
        """these sas7bdat and sasbcat where produced on windows, probably 32 bit"""
        dat = os.path.join(self.catalog_data_folder, "test_data_win.sas7bdat")
        cat = os.path.join(self.catalog_data_folder, "test_formats_win.sas7bcat")
        df, meta = pyreadstat.read_sas7bdat(dat, catalog_file=cat, output_format=self.backend)
        self.assertTrue(df.equals(self.df_sas_format))

    def test_sas_catalog_linux(self):
        """these sas7bdat and sasbcat where produced on linux 64 bit"""
        dat = os.path.join(self.catalog_data_folder, "test_data_linux.sas7bdat")
        cat = os.path.join(self.catalog_data_folder, "test_formats_linux.sas7bcat")
        df, meta = pyreadstat.read_sas7bdat(dat, catalog_file=cat, output_format=self.backend)
        self.assertTrue(df.equals(self.df_sas_format))

    def test_sas_dates(self):
        sas_file = os.path.join(self.basic_data_folder, "dates.sas7bdat")
        df_sas, meta = pyreadstat.read_sas7bdat(sas_file, output_format=self.backend)
        self.assertTrue(df_sas.equals(self.df_sas_dates))

    def test_sas_dates_as_pandas(self):
        sas_file = os.path.join(self.basic_data_folder, "dates.sas7bdat")
        df_sas, meta = pyreadstat.read_sas7bdat(sas_file, dates_as_pandas_datetime=True, output_format=self.backend)
        self.assertTrue(df_sas.equals(self.df_sas_dates_as_pandas))
        


    def test_sas_user_missing(self):
        sas_file = os.path.join(self.missing_data_folder, "missing_test.sas7bdat")
        cat_file = os.path.join(self.missing_data_folder, "missing_formats.sas7bcat")
        unformatted_csv = os.path.join(self.missing_data_folder, "missing_unformatted.csv")
        formatted_csv = os.path.join(self.missing_data_folder, "missing_sas_formatted.csv")
        labeled_csv = os.path.join(self.missing_data_folder, "missing_sas_labeled.csv")
        
        df_sas, meta = pyreadstat.read_sas7bdat(sas_file, output_format=self.backend)
        kwds = {}
        if self.backend == "polars":
            kwds["null_values"] = "NaN"
        df_csv = nw.read_csv(unformatted_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        
        df_sas, meta = pyreadstat.read_sas7bdat(sas_file, user_missing=True, output_format=self.backend)
        df_csv = nw.read_csv(formatted_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        missing_user_values = {'var1':['A'],'var2': ['B'], 'var3':['C'], 'var4':['X'], 'var5':['Y'], 
        'var6':['Z'], 'var7':['_']}
        self.assertDictEqual(meta.missing_user_values, missing_user_values)

        df_sas, meta = pyreadstat.read_sas7bdat(sas_file,
                            catalog_file=cat_file, user_missing=True,
                            formats_as_category=False, output_format=self.backend)
        df_csv = nw.read_csv(labeled_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        
    def test_dta_user_missing(self):
        dta_file = os.path.join(self.missing_data_folder, "missing_test.dta")
        unformatted_csv = os.path.join(self.missing_data_folder, "missing_unformatted.csv")
        formatted_csv = os.path.join(self.missing_data_folder, "missing_dta_formatted.csv")
        labeled_csv = os.path.join(self.missing_data_folder, "missing_dta_labeled.csv")
        
        df_sas, meta = pyreadstat.read_dta(dta_file, output_format=self.backend)
        kwds = {}
        if self.backend == "polars":
            kwds["null_values"] = "NaN"
        df_csv = nw.read_csv(unformatted_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        
        df_sas, meta = pyreadstat.read_dta(dta_file, user_missing=True, output_format=self.backend)
        df_csv = nw.read_csv(formatted_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        missing_user_values = {'var1':['a'],'var2': ['b'], 'var3':['c'], 'var4':['x'], 'var5':['y'], 'var6':['z']}
        self.assertDictEqual(meta.missing_user_values, missing_user_values)
        
        df_sas, meta = pyreadstat.read_dta(dta_file,
                            apply_value_formats=True, user_missing=True,
                            formats_as_category=False, output_format=self.backend)
        df_csv = nw.read_csv(labeled_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))

        
    def test_sav_user_missing(self):
        sav_file = os.path.join(self.missing_data_folder, "missing_test.sav")
        unformatted_csv = os.path.join(self.missing_data_folder, "missing_sav_unformatted.csv")
        formatted_csv = os.path.join(self.missing_data_folder, "missing_sav_formatted.csv")
        labeled_csv = os.path.join(self.missing_data_folder, "missing_sav_labeled.csv")
        
        df_sas, meta = pyreadstat.read_sav(sav_file, output_format=self.backend)
        kwds = {}
        if self.backend == "polars":
            kwds["null_values"] = "NaN"
        df_csv = nw.read_csv(unformatted_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        
        df_sas, meta = pyreadstat.read_sav(sav_file, user_missing=True, output_format=self.backend)
        df_csv = nw.read_csv(formatted_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        
        df_sas, meta = pyreadstat.read_sav(sav_file,
                            apply_value_formats=True, user_missing=True,
                            formats_as_category=False, output_format=self.backend)
        df_sas = nw.from_native(df_sas)
        # in pandas the column will be object with 'missing' and 2.0
        if self.backend == "pandas":
            df_sas = df_sas.with_columns(nw.when(nw.col('var1')==2.0).then(nw.lit('2')).otherwise(nw.col('var1')).alias('var1')).to_native()
        # in polars it will be a string column with 'missing and '2.0'
        else:
            new_ser = nw.new_series(name="var1", values=[str(x) for x in df_sas["var1"]], dtype=nw.String, backend=self.backend)
            df_sas = df_sas.with_columns(new_ser.alias("var1"))
            df_sas = df_sas.with_columns(nw.when(nw.col('var1')=='2.0').then(nw.lit('2')).otherwise(nw.col('var1')).alias('var1')).to_native()
        df_csv = nw.read_csv(labeled_csv, backend=self.backend, **kwds).to_native()
        self.assertTrue(df_sas.equals(df_csv))
        
    def test_sav_missing_char(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.missing_data_folder, "missing_char.sav"), output_format=self.backend)
        if self.backend == "pandas":
            temp = np.nan
        else:
            temp = None
        mdf = nw.from_dict({"mychar": [temp, "a"]}, backend=self.backend).to_native()
        self.assertTrue(df.equals(mdf))
        self.assertTrue(meta.missing_ranges == {})
        df2, meta2 = pyreadstat.read_sav(os.path.join(self.missing_data_folder, "missing_char.sav"), user_missing=True, output_format=self.backend)
        mdf2 = nw.from_dict({"mychar": [ "Z", "a"]}, backend=self.backend).to_native()
        self.assertTrue(df2.equals(mdf2))
        self.assertTrue(meta2.missing_ranges['mychar'][0]=={'lo': "Z", 'hi': "Z"})

    # test reading metadata for multiple response data
    def test_sav_multiple_response(self):
        """Assert MR data is correctly read from sav into metadata."""
        _, meta = pyreadstat.read_sav(os.path.join(self.mr_data_folder, "simple_alltypes.sav"), output_format=self.backend)
        assert meta.mr_sets == {
            "categorical_array": {
                "type": "C",
                "is_dichotomy": False,
                "counted_value": None,
                "label": "",
                "variable_list": ["ca_subvar_1", "ca_subvar_2", "ca_subvar_3"]
            },
            "mymrset": {
                "type": "D",
                "is_dichotomy": True,
                "counted_value": 1,
                "label": "My multiple response set",
                "variable_list": ["bool1", "bool2", "bool3"]
            }
        }
    
    def test_sav_without_multiple_response(self):
        """Assert MR data is read as empty dict when not present in sav."""
        _, meta = pyreadstat.read_sav(os.path.join(self.missing_data_folder, "missing_char.sav"), output_format=self.backend)
        assert meta.mr_sets == {}


    # read in chunks

    def test_chunk_reader(self):
        fpath = os.path.join(self.basic_data_folder, "sample.sas7bdat")
        reader = pyreadstat.read_file_in_chunks(pyreadstat.read_sas7bdat, fpath, chunksize= 2, offset=1, limit=2, disable_datetime_conversion=True, output_format=self.backend)
        
        for df, meta in reader:
            pass
        
        currow = nw.from_native(self.df_nodates_sastata)[1:3].to_native()
        if self.backend == "pandas":
            currow = currow.reset_index(drop=True)
        self.assertTrue(df.equals(currow))

    # read multiprocessing
    def test_multiprocess_reader(self):
        fpath = os.path.join(self.basic_data_folder, "sample_large.sav")
        df_multi, meta_multi = pyreadstat.read_file_multiprocessing(pyreadstat.read_sav, fpath, output_format=self.backend) 
        df_single, meta_single = pyreadstat.read_sav(fpath, output_format=self.backend)
        self.assertTrue(df_multi.equals(df_single))
        self.assertEqual(meta_multi.number_rows, meta_single.number_rows)

    def test_chunk_reader_multiprocess(self):
        fpath = os.path.join(self.basic_data_folder, "sample_large.sav")
        reader = pyreadstat.read_file_in_chunks(pyreadstat.read_sav, fpath, chunksize= 50, multiprocess=True, output_format=self.backend)
        alldfs = list()
        for df, meta in reader:
            alldfs.append(nw.from_native(df))
        df_multi = nw.concat(alldfs, how='vertical').to_native() 
        if self.backend == "pandas":
            df_multi = df_multi.reset_index(drop=True)
        df_single, meta_single = pyreadstat.read_sav(fpath, output_format=self.backend)
        self.assertTrue(df_multi.equals(df_single))

    def test_chunk_reader_multiprocess_dict(self):
        fpath = os.path.join(self.basic_data_folder, "sample_large.sav")
        reader = pyreadstat.read_file_in_chunks(pyreadstat.read_sav, fpath, chunksize= 50, multiprocess=True, output_format='dict')
        alldfs = list()
        for chunkdict, meta in reader:
            if self.backend != "pandas":
                # we need lists not numpy arrays for polars!
                chunkdict = {k:v.tolist() for k,v in chunkdict.items()}
            df = nw.from_dict(chunkdict, backend=self.backend)
            alldfs.append(df)
        df_multi = nw.concat(alldfs, how='vertical').to_native() 
        if self.backend == "pandas":
            df_multi = df_multi.reset_index(drop=True)
        df_single, meta_single = pyreadstat.read_sav(fpath, output_format=self.backend)
        self.assertTrue(df_multi.equals(df_single))

    def test_multiprocess_reader_xport(self):
        fpath = os.path.join(self.basic_data_folder, "sample.xpt")
        df_multi, meta_multi = pyreadstat.read_file_multiprocessing(pyreadstat.read_xport, fpath, num_rows=1000, output_format=self.backend) 
        df_single, meta_single = pyreadstat.read_xport(fpath, output_format=self.backend)
        if is_pandas_version_newer_than_2 and self.backend=="pandas":
            # because in multi there are NaN values in a chunk these columns get object type
            df_single['DTIME'] = df_single['DTIME'].astype(object)
            df_single['MYCHAR'] = df_single['MYCHAR'].astype(object)
        self.assertTrue(df_multi.equals(df_single))


    # writing

    def test_sav_write_basic(self):
        file_label = "basic write"
        file_note = "These are some notes"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        variable_value_labels = {'mylabl': {1.0: 'Male', 2.0: 'Female'}, 'myord': {1.0: 'low', 2.0: 'medium', 3.0: 'high'}}
        missing_ranges = {'mychar':['a'], 'myord': [{'hi':2, 'lo':1}]}
        #variable_alignment = {'mychar':"center", 'myord':"right"}
        variable_display_width = {'mychar':20}
        variable_measure = {"mychar": "nominal"}
        path = os.path.join(self.write_folder, "basic_write.sav")
        pyreadstat.write_sav(self.df_pandas, path, file_label=file_label, column_labels=col_labels, note=file_note, 
            variable_value_labels=variable_value_labels, missing_ranges=missing_ranges, variable_display_width=variable_display_width,
            variable_measure=variable_measure) #, variable_alignment=variable_alignment)
        df, meta = pyreadstat.read_sav(path, user_missing=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        self.assertEqual(meta.notes[0], file_note)
        self.assertDictEqual(meta.variable_value_labels, variable_value_labels)
        self.assertEqual(meta.variable_display_width['mychar'], variable_display_width['mychar'])
        #self.assertDictEqual(meta.variable_alignment, variable_alignment)
        self.assertEqual(meta.variable_measure["mychar"], variable_measure["mychar"])

    def test_sav_write_basic_expanduser(self):
        file_label = "basic write"
        file_note = "These are some notes"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        variable_value_labels = {'mylabl': {1.0: 'Male', 2.0: 'Female'}, 'myord': {1.0: 'low', 2.0: 'medium', 3.0: 'high'}}
        missing_ranges = {'mychar':['a'], 'myord': [{'hi':2, 'lo':1}]}
        #variable_alignment = {'mychar':"center", 'myord':"right"}
        variable_display_width = {'mychar':20}
        variable_measure = {"mychar": "nominal"}
        path = "~/sav_expand.sav" 
        pyreadstat.write_sav(self.df_pandas, path, file_label=file_label, column_labels=col_labels, note=file_note, 
            variable_value_labels=variable_value_labels, missing_ranges=missing_ranges, variable_display_width=variable_display_width,
            variable_measure=variable_measure) #, variable_alignment=variable_alignment)
        df, meta = pyreadstat.read_sav(path, user_missing=True, output_format=self.backend)
        os.remove(os.path.expanduser(path))
        self.assertTrue(df.equals(self.df_pandas))

    def test_zsav_write_basic(self):
        file_label = "basic write"
        file_note = "These are some notes"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        variable_value_labels = {'mylabl': {1.0: 'Male', 2.0: 'Female'}, 'myord': {1.0: 'low', 2.0: 'medium', 3.0: 'high'}}
        missing_ranges = {'mychar':['a'], 'myord': [{'hi':2, 'lo':1}]}
        path = os.path.join(self.write_folder, "basic_write.zsav")
        pyreadstat.write_sav(self.df_pandas, path, file_label=file_label, column_labels=col_labels, compress=True, note=file_note,
                     variable_value_labels=variable_value_labels, missing_ranges=missing_ranges)
        df, meta = pyreadstat.read_sav(path, user_missing=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        self.assertEqual(meta.notes[0], file_note)
        self.assertDictEqual(meta.variable_value_labels, variable_value_labels)

    def test_dta_write_basic(self):
        df_pandas = nw.from_native(self.df_pandas).clone()
        df_pandas = df_pandas.with_columns(nw.col("myord", "mylabl").cast(nw.Int32)).to_native()

        file_label = "basic write"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        variable_value_labels = {'mylabl': {1: 'Male', 2: 'Female'}, 'myord': {1: 'low', 2: 'medium', 3: 'high'}}
        path = os.path.join(self.write_folder, "basic_write.dta")
        pyreadstat.write_dta(df_pandas, path, file_label=file_label, column_labels=col_labels, version=12, variable_value_labels=variable_value_labels)
        df, meta = pyreadstat.read_dta(path, output_format=self.backend)

        df_pandas = nw.from_native(df_pandas).with_columns(nw.col("myord", "mylabl").cast(nw.Int64)).to_native()

        self.assertTrue(df.equals(df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        self.assertDictEqual(meta.variable_value_labels, variable_value_labels)

    def test_dta_write_user_missing(self):
        # bug in nw, nw.Object is not translated to pd.Object
        if self.backend == "pandas":
            df_csv = pd.DataFrame.from_dict({"Var1": [3, "a"], "Var2":["a", "b"]})
            df_csv2 = pd.DataFrame.from_dict({"Var1": [3, "labeled"], "Var2":["a", "b"]})
        else:
            df_csv = nw.from_dict({"Var1": [3, "a"], "Var2":["a", "b"]}, backend=self.backend, schema={"Var1":nw.Object, "Var2":nw.String}).to_native()
            df_csv2 = nw.from_dict({"Var1": [3, "labeled"], "Var2":["a", "b"]}, backend=self.backend, schema={"Var1":nw.Object, "Var2":nw.String}).to_native()

        missing_user_values = {'Var1': ['a']}
        variable_value_labels = {'Var1':{'a':'labeled'}}
        path = os.path.join(self.write_folder, "user_missing_write.dta")
        pyreadstat.write_dta(df_csv, path, version=12, missing_user_values=missing_user_values, variable_value_labels=variable_value_labels)
        
        df_dta, meta = pyreadstat.read_dta(path, user_missing=True, output_format=self.backend)
        # in polars comparing two series of type object always gives error
        if self.backend != "pandas":
            new_ser = nw.new_series(name="Var1", values=[str(x) for x in df_csv["Var1"]], dtype=nw.String, backend=self.backend)
            df_csv = nw.from_native(df_csv).with_columns(new_ser.alias("Var1")).to_native()
            new_ser = nw.new_series(name="Var1", values=[str(x) for x in df_dta["Var1"]], dtype=nw.String, backend=self.backend)
            df_dta = nw.from_native(df_dta).with_columns(new_ser.alias("Var1")).to_native()
            
        self.assertTrue(df_csv.equals(df_dta))
        self.assertDictEqual(meta.missing_user_values, missing_user_values)
        
        df_dta2, meta2 = pyreadstat.read_dta(path, user_missing=True, apply_value_formats=True, formats_as_category=False, output_format=self.backend)
        if self.backend != "pandas":
            new_ser = nw.new_series(name="Var1", values=[str(x) for x in df_csv2["Var1"]], dtype=nw.String, backend=self.backend)
            df_csv2 = nw.from_native(df_csv2).with_columns(new_ser.alias("Var1")).to_native()
            new_ser = nw.new_series(name="Var1", values=[str(x) for x in df_dta2["Var1"]], dtype=nw.String, backend=self.backend)
            df_dta2 = nw.from_native(df_dta2).with_columns(new_ser.alias("Var1")).to_native()
        self.assertTrue(df_csv2.equals(df_dta2))


    def test_xport_write_basic_v8(self):
        file_label = "basic write"
        table_name = "TEST"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        path = os.path.join(self.write_folder, "write.xpt")
        pyreadstat.write_xport(self.df_pandas, path, file_label=file_label, column_labels=col_labels, table_name=table_name, file_format_version=8)
        df, meta = pyreadstat.read_xport(path, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]

        self.assertTrue(df.equals(self.df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        self.assertEqual(table_name, meta.table_name)

    def test_xport_write_basic_v5(self):
        file_label = "basic write"
        table_name = "TEST"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        path = os.path.join(self.write_folder, "write.xpt")
        pyreadstat.write_xport(self.df_pandas, path, file_label=file_label, column_labels=col_labels, table_name=table_name, file_format_version=5)
        df, meta = pyreadstat.read_xport(path, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]

        self.assertTrue(df.equals(self.df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        self.assertEqual(table_name, meta.table_name)

    def test_por_write_basic(self):

        file_label = "basic write"
        #file_note = "These are some notes"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        path = os.path.join(self.write_folder, "write.por")
        pyreadstat.write_por(self.df_pandas, path, file_label=file_label, column_labels=col_labels) #, note=file_note)
        df, meta = pyreadstat.read_por(path, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        self.assertTrue(df.equals(self.df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        #self.assertEqual(meta.notes[0], file_note)


    def test_sav_write_dates(self):

        path = os.path.join(self.write_folder, "dates_write.sav")
        pyreadstat.write_sav(self.df_sas_dates2, path)
        df, meta = pyreadstat.read_sav(path, output_format=self.backend)
        self.assertTrue(df.equals(self.df_sas_dates2))

    def test_zsav_write_dates(self):

        path = os.path.join(self.write_folder, "dates_write_zsav.sav")
        pyreadstat.write_sav(self.df_sas_dates2, path, compress=True)
        df, meta = pyreadstat.read_sav(path, output_format=self.backend)
        self.assertTrue(df.equals(self.df_sas_dates2))

    def test_dta_write_dates(self):

        path = os.path.join(self.write_folder, "dates_write.dta")
        pyreadstat.write_dta(self.df_sas_dates, path)
        df, meta = pyreadstat.read_dta(path, output_format=self.backend)
        self.assertTrue(df.equals(self.df_sas_dates))

    def test_xport_write_dates(self):

        path = os.path.join(self.write_folder, "dates_write.xpt")
        pyreadstat.write_xport(self.df_sas_dates2, path)
        df, meta = pyreadstat.read_xport(path, output_format=self.backend)
        self.assertTrue(df.equals(self.df_sas_dates2))

    def test_sav_write_charnan(self):
        path = os.path.join(self.write_folder, "charnan.sav")
        pyreadstat.write_sav(self.df_charnan, path)
        df, meta = pyreadstat.read_sav(path, output_format=self.backend)
        df2 = nw.from_native(self.df_charnan)
        if self.backend != "pandas":
            new_ser = nw.new_series(name="object", values=[str(x) if x is not None else None for x in df2["object"]], dtype=nw.String, backend=self.backend)
            df2 = df2.with_columns(new_ser.alias("object"))
        df2 = df2.with_columns(nw.col("string", "object").fill_null(""))
        df2 = df2.with_columns(nw.col("integer").cast(nw.Float64))
        if self.backend == "pandas":
            df2 = df2.with_columns(nw.col("object").cast(nw.String))
        df2 = df2.to_native()
        self.assertTrue(df2.equals(df))

    def test_zsav_write_charnan(self):
        path = os.path.join(self.write_folder, "charnan_zsav.sav")
        pyreadstat.write_sav(self.df_charnan, path, compress=True)
        df, meta = pyreadstat.read_sav(path, output_format=self.backend)
        df2 = nw.from_native(self.df_charnan)
        if self.backend != "pandas":
            new_ser = nw.new_series(name="object", values=[str(x) if x is not None else None for x in df2["object"]], dtype=nw.String, backend=self.backend)
            df2 = df2.with_columns(new_ser.alias("object"))
        df2 = df2.with_columns(nw.col("string", "object").fill_null(""))
        df2 = df2.with_columns(nw.col("integer").cast(nw.Float64))
        if self.backend == "pandas":
            df2 = df2.with_columns(nw.col("object").cast(nw.String))
        df2 = df2.to_native()
        self.assertTrue(df2.equals(df))

    def test_xport_write_charnan(self):
        path = os.path.join(self.write_folder, "charnan.xpt")
        pyreadstat.write_xport(self.df_charnan, path)
        df, meta = pyreadstat.read_xport(path, output_format=self.backend)
        df2 = nw.from_native(self.df_charnan)
        if self.backend != "pandas":
            new_ser = nw.new_series(name="object", values=[str(x) if x is not None else None for x in df2["object"]], dtype=nw.String, backend=self.backend)
            df2 = df2.with_columns(new_ser.alias("object"))
        df2 = df2.with_columns(nw.col("string", "object").fill_null(""))
        df2 = df2.with_columns(nw.col("integer").cast(nw.Float64))
        if self.backend == "pandas":
            df2 = df2.with_columns(nw.col("object").cast(nw.String))
        df2 = df2.to_native()
        self.assertTrue(df2.equals(df))

    def test_por_write_charnan(self):
        path = os.path.join(self.write_folder, "charnan_zsav.por")
        pyreadstat.write_por(self.df_charnan, path)
        df, meta = pyreadstat.read_por(path, output_format=self.backend)
        df.columns = [x.lower() for x in df.columns]
        df2 = nw.from_native(self.df_charnan)
        if self.backend != "pandas":
            new_ser = nw.new_series(name="object", values=[str(x) if x is not None else None for x in df2["object"]], dtype=nw.String, backend=self.backend)
            df2 = df2.with_columns(new_ser.alias("object"))
        df2 = df2.with_columns(nw.col("string", "object").fill_null(""))
        df2 = df2.with_columns(nw.col("integer").cast(nw.Float64))
        if self.backend == "pandas":
            df2 = df2.with_columns(nw.col("object").cast(nw.String))
        df2 = df2.to_native()
        self.assertTrue(df2.equals(df))

    def test_dta_write_charnan(self):
        path = os.path.join(self.write_folder, "charnan.dta")
        pyreadstat.write_dta(self.df_charnan, path)
        df, meta = pyreadstat.read_dta(path, output_format=self.backend)
        df2 = nw.from_native(self.df_charnan)
        if self.backend != "pandas":
            new_ser = nw.new_series(name="object", values=[str(x) if x is not None else None for x in df2["object"]], dtype=nw.String, backend=self.backend)
            df2 = df2.with_columns(new_ser.alias("object"))
        df2 = df2.with_columns(nw.col("string", "object").fill_null(""))
        df2 = df2.with_columns(nw.col("integer").cast(nw.Float64))
        if self.backend == "pandas":
            df2 = df2.with_columns(nw.col("object").cast(nw.String))
        df2 = df2.to_native()
        self.assertTrue(df2.equals(df))

    def test_set_value_labels(self):

        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), output_format=self.backend)
        df_formatted = pyreadstat.set_value_labels(df, meta, formats_as_category=True)
        self.assertTrue(df_formatted.equals(self.df_pandas_formatted))
        # partial
        sub1_raw = df[['myord']]
        sub1 = pyreadstat.set_value_labels(sub1_raw, meta, formats_as_category=True)
        sub2 = self.df_pandas_formatted[['myord']]
        self.assertTrue(sub1.equals(sub2))
        
    def test_update_delete_file(self):
    
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), output_format=self.backend)
        dst_path = os.path.join(self.write_folder, "update_test.sav")
        pyreadstat.write_sav(df, dst_path, variable_value_labels = meta.variable_value_labels)
        # update
        meta.variable_value_labels.update({'mylabl':{1.0:"Gents", 2.0:"Ladies"}})
        pyreadstat.write_sav(df, dst_path, variable_value_labels = meta.variable_value_labels)
        df2, meta2 = pyreadstat.read_sav(dst_path, output_format=self.backend)
        self.assertDictEqual(meta2.variable_value_labels, meta.variable_value_labels)
        os.remove(dst_path)

    def test_xport_write_dates2_v8(self):
        # this sas7bdat file has features that are not compatible with v5
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "dates_xpt.sas7bdat"), output_format=self.backend)
        dst_path = os.path.join(self.write_folder, "dates_xptv8.xpt")
        pyreadstat.write_xport(df, dst_path, file_format_version=8)
        df2, meta2 = pyreadstat.read_xport(dst_path, output_format=self.backend)
        self.assertTrue(df.equals(df2))

    def test_xport_dates2_v8(self):
        # this sas7bdat file has features that are not compatible with v5
        df, meta = pyreadstat.read_sas7bdat(os.path.join(self.basic_data_folder, "dates_xpt.sas7bdat"), output_format=self.backend)
        # this xpt file was written in SAS from the sas7bdat file
        df2, meta2 = pyreadstat.read_xport(os.path.join(self.basic_data_folder, "dates_xpt_v8.xpt"), output_format=self.backend)
        self.assertTrue(df.equals(df2))
        self.assertListEqual(meta.column_labels, meta2.column_labels)

    def test_sav_international_utf8_char_value(self):
        # a file that has a value with international characters and the file is coded in utf-8
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "tegulu.sav"), output_format=self.backend)
        self.assertTrue(nw.from_native(df)[0,1] == "నేను గతంలో వాడిన బ")

    def test_sav_international_varname(self):
        # a file with a varname with international characters
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "hebrews.sav"), output_format=self.backend)
        self.assertTrue(df.columns[0] == "ותק_ב")

    def test_sav_original_var_types(self):
        # a file with a varname with international characters
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "test_width.sav"), output_format=self.backend)
        self.assertEqual(meta.original_variable_types['StartDate'],'A1024')
        self.assertEqual(meta.original_variable_types['ResponseId'],'A18')
        self.assertEqual(meta.original_variable_types['Duration__in_seconds_'],'F40.2')
        self.assertEqual(meta.original_variable_types['Finished'],'F1.0')
        self.assertEqual(meta.readstat_variable_types['Finished'],'double')

    def test_sav_write_longstr(self):
        path = os.path.join(self.write_folder, "longstr.sav")
        pyreadstat.write_sav(self.df_longstr, path, variable_display_width={"v1": 1000})
        df, meta = pyreadstat.read_sav(path, output_format=self.backend)
        self.assertTrue(meta.variable_display_width['v1']==1000)
        self.assertTrue(len(nw.from_native(df)[0,0])==2345)

    def test_dta_write_longstr(self):
        path = os.path.join(self.write_folder, "longstr.dta")
        df_longstr = self.df_longstr
        pyreadstat.write_dta(self.df_longstr, path)
        df, meta = pyreadstat.read_dta(path, output_format=self.backend)
        self.assertTrue(df.equals(df_longstr))
        
        
    def test_sas7bdat_file_label_linux(self):
        "testing file label for file produced on linux"
        path = os.path.join(self.basic_data_folder, "test_file_label_linux.sas7bdat")
        df, meta = pyreadstat.read_sas7bdat(path, output_format=self.backend)
        self.assertEqual(meta.file_label, "mytest label")
        self.assertEqual(meta.table_name, "TEST_DATA")

    def test_sas7bdat_extra_date_formats(self):
        "testing extra date format argument"
        path = os.path.join(self.basic_data_folder, "date_test.sas7bdat")
        df, meta = pyreadstat.read_sas7bdat(path, extra_date_formats=["MMYY", "YEAR"], output_format=self.backend)
        self.assertEqual(nw.from_native(df)['yr'][0], date(2023,1,1))
        self.assertEqual(nw.from_native(df)['dtc4'][0], date(2023,7,1))

    def test_sas7bdat_file_label_windows(self):
        "testing file label for file produced on windows"
        path = os.path.join(self.basic_data_folder, "test_file_label_win.sas7bdat")
        df, meta = pyreadstat.read_sas7bdat(path, output_format=self.backend)
        self.assertEqual(meta.file_label, "mytest label")
        self.assertEqual(meta.table_name, "TEST_DATA")

    def test_sav_write_variable_formats(self):
        "testing variable formats for SAV files"
        path = os.path.join(self.write_folder, "variable_format.sav")
        df = nw.from_dict({'restricted':[1023, 10], 'integer':[1,2]}, backend=self.backend)
        formats = {'restricted':'restricted_integer', 'integer':'integer'}
        pyreadstat.write_sav(df, path, variable_format=formats)
        df2, meta2 = pyreadstat.read_sav(path, output_format=self.backend)
        self.assertEqual(meta2.original_variable_types['restricted'], "N4")
        self.assertEqual(meta2.original_variable_types['integer'], "F1.0")

    def test_sav_ordered_categories(self):
        path = os.path.join(self.basic_data_folder, "ordered_category.sav")
        df, meta = pyreadstat.read_sav(path, apply_value_formats=True, formats_as_ordered_category=True, output_format=self.backend)
        # in polars it is enum, it is ordered by default
        if self.backend == "pandas":
            self.assertTrue(df["Col1"].cat.ordered)
        else:
            self.assertTrue(type(nw.from_native(df)["Col1"].dtype)==nw.Enum)
        self.assertListEqual(list(nw.from_native(df)["Col1"].cat.get_categories().to_list()), ['high', 'low', 'medium'])

    def test_sav_pathlib(self):
        if is_pathlib_available:
            path = Path(self.basic_data_folder).joinpath("sample.sav")
            df, meta = pyreadstat.read_sav(path, output_format=self.backend)
            self.assertTrue(df.equals(self.df_pandas))

    def test_sav_write_pathlib(self):
        if is_pathlib_available:
            file_label = "basic write"
            file_note = "These are some notes"
            col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
            variable_value_labels = {'mylabl': {1.0: 'Male', 2.0: 'Female'}, 'myord': {1.0: 'low', 2.0: 'medium', 3.0: 'high'}}
            missing_ranges = {'mychar':['a'], 'myord': [{'hi':2, 'lo':1}]}
            #variable_alignment = {'mychar':"center", 'myord':"right"}
            variable_display_width = {'mychar':20}
            variable_measure = {"mychar": "nominal"}
            path = Path(self.write_folder).joinpath('pathlib_write.sav')
            pyreadstat.write_sav(self.df_pandas, path, file_label=file_label, column_labels=col_labels, note=file_note, 
                variable_value_labels=variable_value_labels, missing_ranges=missing_ranges, variable_display_width=variable_display_width,
                variable_measure=variable_measure) #, variable_alignment=variable_alignment)
            df, meta = pyreadstat.read_sav(path, user_missing=True, output_format=self.backend)
            self.assertTrue(df.equals(self.df_pandas))
            self.assertEqual(meta.file_label, file_label)
            self.assertListEqual(meta.column_labels, col_labels)
            self.assertEqual(meta.notes[0], file_note)
            self.assertDictEqual(meta.variable_value_labels, variable_value_labels)
            self.assertEqual(meta.variable_display_width['mychar'], variable_display_width['mychar'])
            #self.assertDictEqual(meta.variable_alignment, variable_alignment)
            self.assertEqual(meta.variable_measure["mychar"], variable_measure["mychar"])

    def test_sav_write_dictlabels(self):
        col_names = ["mychar", "mynum", "mydate", "dtime", "mylabl", "myord", "mytime"]
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        col_lab_dict = {k:v for k,v in zip(col_names, col_labels) if v}
        variable_value_labels = {'mylabl': {1.0: 'Male', 2.0: 'Female'}, 'myord': {1.0: 'low', 2.0: 'medium', 3.0: 'high'}}
        missing_ranges = {'mychar':['a'], 'myord': [{'hi':2, 'lo':1}]}
        path = os.path.join(self.write_folder, "dictlabel_write.sav")
        pyreadstat.write_sav(self.df_pandas, path, column_labels=col_lab_dict)
        df, meta = pyreadstat.read_sav(path, user_missing=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertListEqual(meta.column_labels, col_labels)
    
    def test_dta_write_single_value_user_missing(self):
        df = nw.from_dict({"var": ["a", "a", "a", "a"]}, backend=self.backend).to_native()
        missing_user_values = {"var": ["a"]}
        path = os.path.join(self.write_folder, "singleusermissing.dta")
        pyreadstat.write_dta(df=df, dst_path=path, missing_user_values=missing_user_values,version=12) 
        df2, meta2 = pyreadstat.read_dta(path, user_missing=True, output_format=self.backend)
        self.assertTrue(df.equals(df2))

    def test_dta_write_only_missing_and_user_missing(self):
        if self.backend=="pandas":
            temp = np.nan
        else:
            temp = None
        df = nw.from_dict({"var": [temp, "a", "b"]}, backend=self.backend).to_native()
        path = os.path.join(self.write_folder, "onlymissing_and_usermissing.dta")
        variable_value_labels={"var": {1: "Val 1", 2: "Val 2", 3: "Val 3", "a": "Missing A", "b": "Missing B", } }
        missing_user_values={"var": ["a", "b"]}
        pyreadstat.write_dta(df, path, variable_value_labels=variable_value_labels, missing_user_values=missing_user_values,version=12)
        df2, meta2 = pyreadstat.read_dta(path, user_missing=True, output_format=self.backend)
        self.assertTrue(df.equals(df2))

    def test_sav_outputformat_dict(self):
        df, meta = pyreadstat.read_sav(os.path.join(self.basic_data_folder, "sample.sav"), output_format='dict')
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(self.df_pandas))
        self.assertTrue(len(meta.notes)>0)
        self.assertTrue(meta.variable_display_width["mychar"]==9)
        self.assertTrue(meta.variable_storage_width["mychar"] == 8)
        self.assertTrue(meta.variable_measure["mychar"]=="nominal")
        self.assertTrue(meta.readstat_variable_types["mychar"]=="string")
        self.assertTrue(meta.readstat_variable_types["myord"]=="double")
        kwds = {}
        if self.backend == "pandas":
            kwds['orient'] = 'list'
        padic = self.df_pandas.to_dict(**kwds)
        for colname, data in df.items():
            curdfcol = df[colname]
            for indx, val in enumerate(data):
                if pd.isna(val) and pd.isna(curdfcol[indx]):
                    continue
                self.assertTrue(val==curdfcol[indx])

    def test_sav_write_rowcompression(self):
        file_label = "row compression write"
        file_note = "These are some notes"
        col_labels = ["mychar label","mynum label", "mydate label", "dtime label", None, "myord label", "mytime label"]
        variable_value_labels = {'mylabl': {1.0: 'Male', 2.0: 'Female'}, 'myord': {1.0: 'low', 2.0: 'medium', 3.0: 'high'}}
        missing_ranges = {'mychar':['a'], 'myord': [{'hi':2, 'lo':1}]}
        #variable_alignment = {'mychar':"center", 'myord':"right"}
        variable_display_width = {'mychar':20}
        variable_measure = {"mychar": "nominal"}
        path = os.path.join(self.write_folder, "rowcompression_write.sav")
        pyreadstat.write_sav(self.df_pandas, path, file_label=file_label, column_labels=col_labels, note=file_note, 
            variable_value_labels=variable_value_labels, missing_ranges=missing_ranges, variable_display_width=variable_display_width,
            variable_measure=variable_measure, row_compress=True) #, variable_alignment=variable_alignment)
        df, meta = pyreadstat.read_sav(path, user_missing=True, output_format=self.backend)
        self.assertTrue(df.equals(self.df_pandas))
        self.assertEqual(meta.file_label, file_label)
        self.assertListEqual(meta.column_labels, col_labels)
        self.assertEqual(meta.notes[0], file_note)
        self.assertDictEqual(meta.variable_value_labels, variable_value_labels)
        self.assertEqual(meta.variable_display_width['mychar'], variable_display_width['mychar'])
        #self.assertDictEqual(meta.variable_alignment, variable_alignment)
        self.assertEqual(meta.variable_measure["mychar"], variable_measure["mychar"])

    def test_dta_write_int_missing(self):
        if self.backend == "pandas":
            df = pd.DataFrame.from_dict({'a': [ 1, 2, None]},dtype='Int32')
        else:
            df = nw.from_dict({'a': [ 1, 2, None]}, backend=self.backend, schema={'a': nw.Int32}).to_native()
        path = os.path.join(self.write_folder, "missingint.dta")
        pyreadstat.write_dta(df, path)
        df2, meta2 = pyreadstat.read_dta(path, output_format=self.backend)
        if self.backend == "pandas":
            df2["a"] = df2["a"].astype('Int32')
        self.assertTrue(df.equals(df2))

    def test_dta_write_bool_missing(self):
        if self.backend == "pandas":
            df = pd.DataFrame.from_dict({'a': [ True, False, None]},dtype='boolean')
        else:
            df = nw.from_dict({'a': [ True, False, None]}, backend=self.backend, schema={'a': nw.Boolean}).to_native()
        path = os.path.join(self.write_folder, "missingbool.dta")
        pyreadstat.write_dta(df, path)
        df2, meta2 = pyreadstat.read_dta(path, output_format=self.backend)
        if self.backend == "pandas":
            df2["a"] = df2["a"].astype('boolean')
        else:
            df2 = nw.from_native(df2).with_columns(nw.col("a").cast(nw.Boolean)).to_native()
        self.assertTrue(df.equals(df2))

    def test_sav_write_formatted(self):
        path = os.path.join(self.write_folder, "formatted.sav")
        pyreadstat.write_sav(self.df_pandas_formatted, path)
        df, meta = pyreadstat.read_sav(path,  output_format=self.backend)
        df2 = nw.from_native(self.df_pandas_formatted)
        df2 = df2.with_columns(nw.col("myord", "mylabl").cast(nw.String)).to_native()
        self.assertTrue(df.equals(df2))

    def test_sav_write_formatted_nulls(self):
        """
        apply format when some columns have some nulls or all nulls
        all nulls raises a warning in polars
        """
        labels = {'a': {1:'a', 2: 'b'}, 'b': {1:'c'}, 'c': {1:'d'}}
        data = {'a': [1,2,3], 'b':[1, 1, None], 'c': [None, None, None]}
        df = nw.from_dict(data, schema={'a':nw.Float64, 'b':nw.Float64, 'c': nw.Float64}, backend=self.backend).to_native()
        path = os.path.join(self.write_folder, "formatted_nulls.sav")
        pyreadstat.write_sav(df, path, variable_value_labels=labels) 
        schema = dict()
        null = np.nan
        if self.backend != "pandas":
            schema={'a':nw.String, 'b':nw.Categorical, 'c': None}
            null = None
        data_result = {'a': ['a','b', '3.0'], 'b':['c', 'c', null], 'c': [null, null, null]}
        df_control = nw.from_dict(data_result, schema=schema, backend=self.backend).to_native()
        df2, meta2 = pyreadstat.read_sav(path, apply_value_formats=True, output_format=self.backend)
        new_ser = nw.new_series(name="a", values=[str(x) for x in df2["a"]], dtype=nw.String, backend=self.backend)
        df2 = nw.from_native(df2).with_columns(new_ser.alias("a")).to_native()
        cat_type = object
        if is_pandas_version_newer_than_2:
            cat_type = str
        if self.backend == "pandas":
            df_control = nw.from_native(df_control).with_columns(nw.col('b').cast(nw.Categorical)).to_native()
            df_control['c'] = df_control['c'].astype(cat_type).astype("category")
        self.assertTrue(df2.equals(df_control))

    def test_sav_write_mixed_types(self):
        csv_path = os.path.join(self.basic_data_folder, "mixed_types.csv")
        kwds = {}
        if backend == "polars":
            kwds["try_parse_dates"] = True
        df_ori = nw.read_csv(csv_path, backend=self.backend, **kwds)
        df = df_ori.clone()
        time_unit = "ns"
        if is_pandas_version_newer_than_2:
            time_unit = "us"
        if self.backend != "pandas":
            # convince polars to take columns with mixed types
            temp = [0 if x=="0" else x for x in df["col_other"]]
            newser = nw.new_series(name="col_other", values=temp , dtype=nw.Object, backend=df.implementation) 
            df = df.with_columns(newser.alias("col_other"))
            temp = [datetime.strptime('2814-01-01 00:00:00', '%Y-%m-%d %H:%M:%S') if x.startswith('2814') else x for x in df["total_time"]]
            newser = nw.new_series(name="total_time", values=temp, dtype=nw.Object, backend=df.implementation) 
            df = df.with_columns(newser.alias("total_time"))
        else:
            df = df.with_columns(nw.when(nw.col("col_other")=='0').then(nw.lit(0)).otherwise(nw.col('col_other')).alias('col_other'))
            df = df.with_columns(nw.when(nw.col("total_time")=='2814-01-01 00:00:00')
                                   .then(nw.lit(datetime.strptime('2814-01-01 00:00:00', '%Y-%m-%d %H:%M:%S')))
                                   .otherwise(nw.col('total_time'))
                                   .alias('total_time'))
            df = df.with_columns(nw.col("datestamp").cast(nw.Datetime(time_unit=time_unit)))
            df_ori = df_ori.with_columns(nw.col("datestamp").cast(nw.Datetime(time_unit=time_unit)))
        df = df.to_native()
        path = os.path.join(self.write_folder, "mixed_types.sav")
        pyreadstat.write_sav(df, path)
        df2, meta = pyreadstat.read_sav(path,  output_format=self.backend)
        self.assertTrue(df_ori.to_native().equals(df2))

    def test_read_sav_file_handle(self):
        """Test reading SAV file from file-like object (e.g., zip archive)"""
        sav_file = os.path.join(self.basic_data_folder, "sample.sav")
        
        #with tempfile.NamedTemporaryFile(suffix=".zip", delete=True, dir=self.write_folder) as tmp:
        with tempfile.TemporaryDirectory() as tmp_dir:
            zip_file = os.path.join(tmp_dir, "sample.zip")
            with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zf:
                zf.write(sav_file, "sample.sav")
            
            #with zipfile.ZipFile(tmp.name, "r") as zf:
            with zipfile.ZipFile(zip_file, "r") as zf:
                with zf.open("sample.sav", "r") as fh:
                    df, meta = pyreadstat.read_sav(fh, output_format=self.backend)
                    
                    self.assertEqual(len(df.columns), len(self.df_pandas.columns))
                    self.assertEqual(len(df), len(self.df_pandas))

    def test_read_sav_bytesio(self):
        """Test reading SAV file from BytesIO (simulates remote/streaming data)"""
        sav_file = os.path.join(self.basic_data_folder, "sample.sav")
        
        with open(sav_file, "rb") as f:
            file_bytes = f.read()
        
        buffer = io.BytesIO(file_bytes)
        df, meta = pyreadstat.read_sav(buffer, output_format=self.backend)
        
        self.assertEqual(len(df.columns), len(self.df_pandas.columns))
        self.assertEqual(len(df), len(self.df_pandas))
        self.assertListEqual(list(df.columns), list(self.df_pandas.columns))

    def test_sas7bdat_bytesio_chunk(self):
        sas7bdat_file = os.path.join(self.basic_data_folder, "sample.sas7bdat")

        with open(sas7bdat_file, "rb") as f:
            file_bytes = f.read()
        
        buffer = io.BytesIO(file_bytes)

        df, meta = pyreadstat.read_sas7bdat(buffer, row_limit = 2, row_offset =1, output_format=self.backend)
        df_pandas = nw.from_native(self.df_pandas)[1:3].to_native()
        if self.backend == "pandas":
            df_pandas = df_pandas.reset_index(drop=True)
            df_pandas["dtime"] = pd.to_datetime(df_pandas["dtime"])
        self.assertTrue(df.equals(df_pandas))
        self.assertTrue(meta.number_columns == len(self.df_pandas.columns))
        self.assertTrue(meta.number_rows == len(df_pandas))

    def test_multiprocess_reader_bytesio(self):
        fpath = os.path.join(self.basic_data_folder, "sample_large.sav")
        with open(fpath, "rb") as f:
            file_bytes = f.read()
        
        buffer = io.BytesIO(file_bytes)
        df_multi, meta_multi = pyreadstat.read_file_multiprocessing(pyreadstat.read_sav, buffer, output_format=self.backend) 
        df_single, meta_single = pyreadstat.read_sav(fpath, output_format=self.backend)
        self.assertTrue(df_multi.equals(df_single))
        self.assertEqual(meta_multi.number_rows, meta_single.number_rows)


if __name__ == '__main__':

    import sys

    if "--inplace" in sys.argv:
        script_folder = os.path.split(os.path.split(os.path.realpath(__file__))[0])[0]
        sys.path.insert(0, script_folder)
        sys.argv.remove('--inplace')

    backend = "pandas"
    for arg in sys.argv:
        if arg.startswith("--backend"):
            backend = arg.split("=")[1]
            sys.argv.remove(arg)
    # to avoid os.fork warnings when using polars
    #mp.set_start_method('forkserver') can also be 'spawn'
    # very slow, prefer to get the warning
    # TODO: document this in Readme:
    # DeprecationWarning: This process (pid=2314) is multi-threaded, use of fork() may lead to deadlocks in the child self.pid = os.fork()

    print(f"Using backend: {backend}")

    import pyreadstat

    print("package location:", pyreadstat.__file__)

    unittest.main()



# cython: c_string_type=unicode, c_string_encoding=utf8, language_level=2
# ###cython: profile=True
# #############################################################################
# Copyright 2018 Hoffmann-La Roche
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# #############################################################################
import os
import warnings
import sys
import datetime
#import calendar
from datetime import timezone
#from datetime import timezone as _timezone
#from libc.math cimport round, NAN

import numpy as np
import narwhals.stable.v2 as nw

from readstat_api cimport *
from _readstat_parser import ReadstatError, PyreadstatError
from _readstat_parser cimport check_exit_status

cdef set int_types = {nw.Int32, nw.Int16, nw.Int8, nw.UInt16, nw.UInt8, }
cdef set float_types = {nw.Float64, nw.Float32, nw.Decimal, nw.Int128, nw.Int64, nw.UInt128, nw.UInt64, nw.UInt32}
cdef set nat_types = {datetime.datetime, np.datetime64, datetime.time, datetime.date} #pd._libs.tslibs.timestamps.Timestamp,
cdef set pyrwriter_datetimelike_types = {PYWRITER_DATE, PYWRITER_DATETIME, PYWRITER_TIME, PYWRITER_DATETIME64, PYWRITER_DATE64, PYWRITER_TIME64}
cdef set pywriter_numeric_types = {PYWRITER_DOUBLE, PYWRITER_INTEGER, PYWRITER_LOGICAL, PYWRITER_DATE, PYWRITER_DATETIME, PYWRITER_TIME, 
                                   PYWRITER_DATETIME64, PYWRITER_DATE64, PYWRITER_TIME64 }
cdef dict narwhals_to_readstat_types = {PYWRITER_DOUBLE: READSTAT_TYPE_DOUBLE, PYWRITER_INTEGER: READSTAT_TYPE_INT32,
                                      PYWRITER_CHARACTER: READSTAT_TYPE_STRING, PYWRITER_LOGICAL: READSTAT_TYPE_INT32,
                                      PYWRITER_DTA_STR_REF: READSTAT_TYPE_STRING_REF,
                                      PYWRITER_OBJECT: READSTAT_TYPE_STRING, PYWRITER_DATE: READSTAT_TYPE_DOUBLE,
                                      PYWRITER_DATETIME: READSTAT_TYPE_DOUBLE, PYWRITER_TIME: READSTAT_TYPE_DOUBLE,
                                      PYWRITER_DATETIME64: READSTAT_TYPE_DOUBLE, PYWRITER_DATE64: READSTAT_TYPE_DOUBLE, 
                                      PYWRITER_TIME64: READSTAT_TYPE_DOUBLE, }

cdef double spss_offset_secs = 12219379200
cdef double sas_offset_secs = 315619200
cdef double spss_offset_days = 141428
cdef double sas_offset_days = 3653
cdef object date_0 = datetime.datetime(1970,1,1).date()

cdef valid_user_missing_sas = [chr(x) for x in range(ord("A"), ord("Z")+1)] + ["_"]
cdef valid_user_missing_stata = [chr(x) for x in range(ord("a"), ord("z")+1)]

# max lenght of a string in dta before it has to use string_ref
cdef int dta_old_max_width = 128
cdef int dta_111_max_width = 244
cdef int dta_117_max_width = 2045

cdef object vectorized_convert_datetime_to_number(object df, dst_file_format file_format, list pywriter_types, list pywriter_timeunits, int col_count):
    """
    transforms datetime64 columns in the dataframe to floats
    """
    cdef dict convfacs
    cdef double offset_secs
    cdef double mulfac = 1.0
    cdef int col_indx
    cdef list col_indxs
    cdef double convfac

    if file_format == FILE_FORMAT_SAV or file_format == FILE_FORMAT_POR:
        offset_secs = spss_offset_secs
    else:
        offset_secs = sas_offset_secs
    if file_format == FILE_FORMAT_DTA:
        # stata stores in milliseconds
        mulfac = 1000.0
    convfacs = {'ns': 1e9, 'us': 1e6, 'ms': 1e3}

    col_indxs = list()
    for col_indx in range(col_count):
        if pywriter_types[col_indx] == PYWRITER_DATETIME64: 
            col_indxs.append(col_indx)

    df = df.with_columns(nw.nth(col_indxs).cast(nw.Int64))
    for col_indx in col_indxs:
        convfac = convfacs[pywriter_timeunits[col_indx]]
        df = df.with_columns(nw.when(nw.nth(col_indx)!=-9223372036854775808).then(nw.nth(col_indx)))
        df = df.with_columns((((nw.nth(col_indx).cast(nw.Float64))/convfac) + offset_secs).round() * mulfac)
    return df


cdef object vectorized_convert_date_to_number(object df, dst_file_format file_format, list pywriter_types,  int col_count):
    """
    transforms date64 columns in the dataframe to floats
    """
    cdef double offset_days
    cdef double mulfac = 1.0
    cdef int col_indx
    cdef list col_indxs

    if file_format == FILE_FORMAT_SAV or file_format == FILE_FORMAT_POR:
        offset_days = spss_offset_days
        # spss stores in seconds
        mulfac = 86400
    else:
        offset_days = sas_offset_days
    col_indxs = list()
    for col_indx in range(col_count):
        if pywriter_types[col_indx] == PYWRITER_DATE64: 
            col_indxs.append(col_indx)

    df = df.with_columns(nw.nth(col_indxs).cast(nw.Int64))
    for col_indx in col_indxs:
        df = df.with_columns(nw.when(nw.nth(col_indx)!=-9223372036854775808).then(nw.nth(col_indx)))
        df = df.with_columns((nw.nth(col_indx).cast(nw.Float64) + offset_days).round() * mulfac)
    return df

cdef object vectorized_convert_time_to_number(object df, dst_file_format file_format, list pywriter_types,  int col_count):
    """
    transforms time64 columns in the dataframe to floats
    """
    cdef double mulfac = 1.0
    cdef int col_indx
    cdef list col_indxs

    if file_format == FILE_FORMAT_DTA:
        # stata stores in milliseconds
        mulfac = 1000.0

    col_indxs = list()
    for col_indx in range(col_count):
        if pywriter_types[col_indx] == PYWRITER_TIME64: 
            col_indxs.append(col_indx)

    df = df.with_columns(nw.nth(col_indxs).cast(nw.Int64))
    for col_indx in col_indxs:
        df = df.with_columns(nw.when(nw.nth(col_indx)!=-9223372036854775808).then(nw.nth(col_indx)))
        df = df.with_columns((nw.nth(col_indx).cast(nw.Float64)/1e9).round() * mulfac)
    return df

cdef double convert_datetimelike_to_number(dst_file_format file_format, pywriter_variable_type curtype, object curval) except *:
    """
    converts a datime like python/pandas object to a float
    """

    cdef double offset_days, tstamp

    if file_format == FILE_FORMAT_SAV or file_format == FILE_FORMAT_POR:
        offset_days = spss_offset_days
        offset_secs = spss_offset_secs
    else:
        offset_days = sas_offset_days
        offset_secs = sas_offset_secs

    if curtype == PYWRITER_DATETIME or curtype == PYWRITER_DATETIME64:
        # get timestamp in seconds
        if type(curval) == datetime.datetime:
            tstamp = curval.replace(tzinfo=timezone.utc).timestamp() # works only in python 3
            #tstamp = calendar.timegm(curval.replace(tzinfo=_timezone.utc).timetuple())

        tstamp += offset_secs
        if file_format == FILE_FORMAT_DTA:
            # stata stores in milliseconds
            tstamp *= 1000

    elif curtype == PYWRITER_DATE:
        if type(curval) == datetime.date:
            days = curval - date_0
            tstamp = days.days
            tstamp += offset_days
            if file_format == FILE_FORMAT_SAV or file_format == FILE_FORMAT_POR:
                # spss stores in seconds
                tstamp *= 86400

    elif curtype == PYWRITER_TIME:
        if type(curval) == datetime.time:
            tdelta = datetime.datetime.combine(datetime.date.min, curval) - datetime.datetime.min
            tstamp = tdelta.total_seconds()
            #tstamp += offset * 86400
            if file_format == FILE_FORMAT_DTA:
                # stata stores in milliseconds
                tstamp *= 1000

    return tstamp

cdef char * get_datetimelike_format_for_readstat(dst_file_format file_format, pywriter_variable_type curtype):
    """
    gives back a string with the format of the variable (according to the final application) to be used by readstat 
    """

    if curtype == PYWRITER_DATE or curtype == PYWRITER_DATE64:
        if file_format == FILE_FORMAT_DTA:
            return "%td"
        #elif file_format == FILE_FORMAT_SAV:
        #    return "DATE11"
        else:
            return "DATE"
    elif curtype == PYWRITER_DATETIME or curtype == PYWRITER_DATETIME64:
        if file_format == FILE_FORMAT_DTA:
            return "%tc"
        #elif file_format == FILE_FORMAT_SAV:
        #    return "DATETIME20"
        else:
            return "DATETIME"
    elif curtype == PYWRITER_TIME or curtype == PYWRITER_TIME64:
        if file_format == FILE_FORMAT_DTA:
            return "%tcHH:MM:SS"
        else:
            return "TIME"
    else:
        raise PyreadstatError("Unknown pywriter variable format")

cdef int get_narwhals_str_series_max_length(object series, dict value_labels, bint isobject):
    """ For a string series get the max length of the strings. Assumes there is no NaN among the elements. 
    """
    cdef object val
    cdef bytes temp
    cdef int max_length = 1
    cdef int curlen
    cdef list labels
    for val in series:
        if isobject:
            val = str(val)
        temp = val.encode("utf-8")
        curlen = len(temp)
        if curlen > max_length:
            max_length = curlen
    if value_labels:
        labels = list(value_labels.keys())
        for lab in labels:
            curlen = len(str(lab))
            if curlen > max_length:
                max_length = curlen

    return max_length


cdef int check_series_all_same_types(object series, object type_to_check):
    """
    1 if all elements in a series are of type type_to_check, 0 otherwise
    """
    for val in series:
        if type(val) != type_to_check:
            return 0
    return 1

cdef list get_narwhals_column_types(object df, dict missing_user_values, dict variable_value_labels, int dta_str_max_len):
    """
    From a narwhals data frame, get a list with tuples column types as first element, max_length as second, is_missing
    as third and time units as fourth.
    max_lenght is the max length of a string or string representation of an object, 0 for numeric types. is_missing flags
    wether the series has missing values (1) or not (0)
    dta_str_max_len is the max length for a dta string, 0 if the file format is not dta
    """

    cdef int max_length, isobject
    cdef bint has_missing
    cdef list result = list()
    cdef int equal, is_missing

    if variable_value_labels is None:
        variable_value_labels = dict()

    for curseries in df.iter_columns():
        col_name = curseries.name
        col_type = curseries.dtype

        # if categorical, let's use the type of the categories
        if col_type == nw.Categorical:
            col_type = curseries.cat.get_categories().dtype

        # we need to remove nulls for series inspection
        has_missing = <bint>curseries.null_count()>0
        if has_missing:
            curseries = curseries.drop_nulls()
        
        # missing_user_values are strings that represent a missing value in a numeric column
        # they are object because it is a mix of string and number.
        # it could also be that the type of the series is string because it contains only
        # missing values and missing_user_values.
        # we need to take out those before inspecting the object series futher
        curuser_missing = None
        if missing_user_values:
            curuser_missing = missing_user_values.get(col_name)
        if curuser_missing:
            if not df.implementation.is_pandas() and col_type == nw.Object:
                curseries = [x for x in curseries if x not in curuser_missing]
            else:
                curseries = curseries.filter(~curseries.is_in(curuser_missing))
            if not len(curseries):
                result.append((PYWRITER_DOUBLE, 0, 1, None))
                continue
        
        max_length = 0
        curtype = None
        equal = True
        # let's deal first with object type, Enum could also be anything
        if col_type == nw.Object or col_type==nw.Enum:
            curtype = type(curseries[0])
            equal = check_series_all_same_types(curseries, curtype)
            # if all elements are equal, they could be a few we expect to be an object class
            # or it could be that they are some other common types (numeric) after removing the missing_user_values
            # therefore if one of these conditions are not met they continuing flowing into the next if
            # if not all the elements of the series are from the same type, then we deal with it at the end
            # as other types we don't know what to do with
            if equal:
                # types expected to be object
                if curtype == datetime.date:
                    result.append((PYWRITER_DATE, 0, has_missing, None))
                    continue
                elif curtype == datetime.datetime:
                    result.append((PYWRITER_DATETIME, 0, has_missing, None))
                    continue
                elif curtype == datetime.time:
                    result.append((PYWRITER_TIME, 0, has_missing, None))
                    continue

        # numeric types: they could contain missing_user_values
        if equal and (col_type in float_types or curtype == float):
            result.append((PYWRITER_DOUBLE, 0, has_missing, None))
            continue
        elif equal and (col_type in int_types or curtype == int):
            result.append((PYWRITER_INTEGER, 0,has_missing, None))
            continue
        elif equal and (col_type == nw.Boolean or curtype == bool):
            result.append((PYWRITER_LOGICAL, 0,has_missing, None))
            continue
        # these types here should not contain missing_user_values,
        # for string we still check, as later we will raise an error
        elif col_type == nw.String or curtype == str:
            isobject = 0
            if not equal:
                isobject = 1
            max_length = get_narwhals_str_series_max_length(curseries, variable_value_labels.get(col_name), isobject)
            if dta_str_max_len and max_length >= dta_str_max_len:
                result.append((PYWRITER_DTA_STR_REF, max_length, has_missing, None))
                continue
            else:
                if isobject:
                    result.append((PYWRITER_OBJECT, max_length, has_missing, None))
                else:
                    result.append((PYWRITER_CHARACTER, max_length, has_missing, None))
                continue
        elif col_type == nw.Datetime:
            if col_type.time_unit == 'us':
                result.append((PYWRITER_DATETIME64, 0,has_missing, 'us'))
                continue
            elif col_type.time_unit == 'ns':
                result.append((PYWRITER_DATETIME64, 0,has_missing, 'ns'))
                continue
            elif col_type.time_unit == 'ms':
                result.append((PYWRITER_DATETIME64, 0,has_missing, 'ms'))
                continue
            else:
                result.append((PYWRITER_DATETIME, 0, has_missing, None))
                continue
        elif col_type == nw.Date:
            result.append((PYWRITER_DATE64, 0, has_missing, None))
            continue
        elif col_type == nw.Time:
            result.append((PYWRITER_TIME64, 0, has_missing, None))
            continue

        # if the object was not captured by any of the previous cases, we transform it to string
        max_length = get_narwhals_str_series_max_length(curseries, variable_value_labels.get(col_name), 1)
        if dta_str_max_len and max_length >= dta_str_max_len:
            result.append((PYWRITER_DTA_STR_REF, max_length, has_missing, None))
        else:
            result.append((PYWRITER_OBJECT, max_length, has_missing, None))
        continue

    return result

cdef readstat_label_set_t *set_value_label(readstat_writer_t *writer, dict value_labels, str labelset_name,
                        pywriter_variable_type curpytype, dst_file_format file_format, str variable_name, 
                        list user_missing_tags) except *:
    """
    Sets value labels for normal values and also tagged missing values (user defined missing for stata and sas)
    """

    cdef readstat_label_set_t *label_set
    cdef readstat_type_t curtype
    cdef double double_val
    
    curtype = narwhals_to_readstat_types[curpytype]
    label_set = readstat_add_label_set(writer, curtype, labelset_name.encode("utf-8"))

    for value, label in value_labels.items():

        if type(label) != str:
            msg = "variable_value_labels: type of Label %s in variable %s must be string" % (str(label), variable_name)
            raise PyreadstatError(msg)

        if user_missing_tags and value in user_missing_tags:
            if curpytype == PYWRITER_CHARACTER or curpytype == PYWRITER_OBJECT or curpytype==PYWRITER_DTA_STR_REF:
                msg = "missing_user_values not allowed for character variable %s" % variable_name
                raise PyreadstatError(msg)
            
            readstat_label_tagged_value(label_set, ord(value), label.encode("utf-8"))
            continue


        if curpytype == PYWRITER_DOUBLE:
            if type(value) != float and type(value) != int:
                msg = "variable_value_labels: type of Value %s in variable %s must be numeric" % (str(value), variable_name)
                raise PyreadstatError(msg)
            readstat_label_double_value(label_set, value, label.encode("utf-8"))

        elif curpytype == PYWRITER_INTEGER:
            if type(value) != int:
                #if type(value) not in int_types:
                msg = "variable_value_labels: type of Value %s in variable %s must be int" % (str(value), variable_name)
                raise PyreadstatError(msg)
            readstat_label_int32_value(label_set, value, label.encode("utf-8"))

        elif curpytype == PYWRITER_LOGICAL:
            if type(value) != bool and (value != 0 and value != 1):
                msg = "variable_value_labels: type of Value %s in variable %s must be boolean or be 1 or 0" % (str(value), variable_name)
                raise PyreadstatError(msg)
            readstat_label_int32_value(label_set, int(value), label.encode("utf-8"))

        elif curpytype == PYWRITER_CHARACTER or curpytype == PYWRITER_OBJECT or curpytype==PYWRITER_DTA_STR_REF:
            value = str(value)
            readstat_label_string_value(label_set, value.encode("utf-8"), label.encode("utf-8"))

        elif curpytype in (PYWRITER_DATE, PYWRITER_DATETIME, PYWRITER_TIME, PYWRITER_DATETIME64, PYWRITER_DATE64, PYWRITER_TIME64):
            if type(value) not in nat_types:
                msg = "variable_value_labels: type of Value %s in variable %s must match the type of the column in dataframe and be of type date, datetime or time" % (str(value), variable_name)
                raise PyreadstatError(msg)
            double_val = convert_datetimelike_to_number(file_format, curpytype, value) 
            readstat_label_double_value(label_set, double_val, label.encode("utf-8"))

    return label_set

cdef void add_missing_ranges(list cur_ranges, readstat_variable_t *variable, pywriter_variable_type vartype, str variablename) except *:
    """
    Adding missing ranges, user defined missing discrete values both numeric and character,
     this happens for SPSS
    """

    cdef int range_values = 0
    cdef int discrete_values = 0
    cdef int discrete_strings = 0

    for cur_range in cur_ranges:
        if isinstance(cur_range, dict):
            hi = cur_range.get("hi")
            lo = cur_range.get("lo")
            if hi is None or lo is None:
                msg = "dictionaries in missing_ranges must have the keys hi and lo"
                raise PyreadstatError(msg)
            if type(hi) in (int, float)  and type(lo) in (int, float):
                if vartype not in pywriter_numeric_types:
                    msg = "numeric missing_ranges value given for non numeric variable %s" %variablename
                    raise PyreadstatError(msg)
                if hi == lo:
                    check_exit_status(readstat_variable_add_missing_double_value(variable, hi))
                    discrete_values += 1
                else:
                    check_exit_status(readstat_variable_add_missing_double_range(variable, lo, hi))
                    range_values += 1
            elif type(hi) == str and type(lo) == str:
                if vartype != PYWRITER_CHARACTER and vartype != PYWRITER_OBJECT and vartype !=PYWRITER_DTA_STR_REF:
                    msg = "character missing_ranges value given for non character variable %s" %variablename
                    raise PyreadstatError(msg)
                if hi == lo:
                    if len(hi) > 8:
                        msg = "missing_ranges: string values length must not be larger than 8"
                        raise PyreadstatError(msg)
                    check_exit_status(readstat_variable_add_missing_string_value(variable, hi))#.encode("utf-8")))
                    discrete_strings += 1
                else:
                    #check_exit_status(readstat_variable_add_missing_string_range(variable, lo, hi))
                    msg = "missing_ranges: hi and lo values must be both the same for string type"
                    raise PyreadstatError(msg)
            else:
                msg = "missing_ranges: hi and lo values must be both either of numeric or string type"
                raise PyreadstatError(msg)
        else:
            if type(cur_range) in (int, float):
                if vartype not in pywriter_numeric_types:
                    msg = "numeric missing_ranges value given for non numeric variable %s" %variablename
                    raise PyreadstatError(msg)
                check_exit_status(readstat_variable_add_missing_double_value(variable, cur_range))
                discrete_values += 1
            elif type(cur_range) == str:
                if vartype != PYWRITER_CHARACTER and vartype != PYWRITER_OBJECT or vartype==PYWRITER_DTA_STR_REF:
                    msg = "character missing_ranges value given for non character variable %s" %variablename
                    raise PyreadstatError(msg)
                if len(cur_range) > 8:
                        msg = "missing_ranges: string values length must not be larger than 8"
                        raise PyreadstatError(msg)
                check_exit_status(readstat_variable_add_missing_string_value(variable, cur_range))#.encode("utf-8")))
                discrete_strings += 1
            else:
                msg = "missing_ranges: values must be both either of numeric or string type"
                raise PyreadstatError(msg)
               
        if discrete_strings > 3:
            msg = "missing_ranges: max 3 string values per variable allowed"
            raise PyreadstatError(msg)
        if range_values:
            if range_values > 1:
                msg = "missing_ranges: max 1 range value per variable allowed"
                raise PyreadstatError(msg)
            if discrete_values > 1:
                msg = "missing_ranges: max 1 discrete numeric value if combined with 1 range value per variable allowed"
                raise PyreadstatError(msg)
        if discrete_values >3:
            msg = "missing_ranges: max 3 discrete numeric values per variable allowed"
            raise PyreadstatError(msg)

cdef void set_variable_alignment(readstat_variable_t *variable, str alignment_str, str var_name) except *:
    """
    Sets the variable alignment, ineffective on SPSS, STATA and XPORT (what about SAS7bdat?)
    """

    cdef readstat_alignment_t alignment

    if alignment_str == "right":
        alignment = READSTAT_ALIGNMENT_RIGHT
    elif alignment_str == "left":
        alignment = READSTAT_ALIGNMENT_LEFT
    elif alignment_str == "center":
        alignment = READSTAT_ALIGNMENT_CENTER
    elif alignment_str == "unknown":
        alignment = READSTAT_ALIGNMENT_UNKNOWN
    else:
        msg = "alignment for variable %s must be either right, center, left or unknown got %s instead" % (var_name, alignment_str)
        raise PyreadstatError(msg)

    readstat_variable_set_alignment(variable, alignment)

cdef void set_variable_display_width(readstat_variable_t *variable, int display_width, str var_name) except *:
    """
    Sets the variable display width (SPSS). Not effective on STATA. (what about SAS7BDAT?)
    """

    readstat_variable_set_display_width(variable, display_width)

cdef void set_variable_measure(readstat_variable_t *variable, str measure_str, str var_name) except *:
    """
    sets the variable measure type (SPSS). Not effective on STATA.
    """

    cdef readstat_measure_t measure

    if measure_str == "nominal":
        measure = READSTAT_MEASURE_NOMINAL
    elif measure_str == "ordinal":
        measure = READSTAT_MEASURE_ORDINAL
    elif measure_str == "scale":
        measure = READSTAT_MEASURE_SCALE
    elif measure_str == "unknown":
        measure = READSTAT_MEASURE_UNKNOWN
    else:
        msg = "measure for variable %s must be either nominal, ordinal, scale or unknown got %s instead" % (var_name, measure_str)
        raise PyreadstatError(msg)

    readstat_variable_set_measure(variable, measure);


cdef ssize_t write_bytes(const void *data, size_t _len, void *ctx) noexcept:
    """
    for the writer an explicit function to write must be defined 
    """
    cdef int fd
    fd = (<int *>ctx)[0]
    if os.name=='nt':
        return _write(fd, data, _len)
    else:
        return write(fd, data, _len)

cdef void _check_exit_status(readstat_error_t retcode) except *:
    """
    transforms a readstat exit status to a python error if status is not READSTAT OK
    """

    cdef char * err_readstat
    cdef str err_message
    if retcode != READSTAT_OK:
        err_readstat = readstat_error_message(retcode)
        err_message = <str> err_readstat
        raise ReadstatError(err_message)

cdef int open_file(bytes filename_bytes):

    cdef int fd
    cdef int flags
    cdef Py_ssize_t length

    cdef char *path

    if os.name == "nt":
        filename_str = os.fsdecode(filename_bytes)
        u16_path = PyUnicode_AsWideCharString(filename_str, &length)
        flags = _O_WRONLY | _O_CREAT | _O_BINARY | _O_TRUNC
        fd = _wsopen(u16_path, flags, _SH_DENYRW, _S_IREAD | _S_IWRITE)
    else:
        path = <char *> filename_bytes
        flags = O_WRONLY | O_CREAT | O_TRUNC
        fd = open(path, flags, 0644)

    return fd

cdef int close_file(int fd):
    if os.name == "nt":
        return _close(fd)
    else:
        return close(fd)

cdef void initial_checks(bint is_pandas, bint is_polars, dict variable_value_labels, dict missing_user_values,
                        dst_file_format file_format, list col_names, bytes filename_bytes) except *:
    """
    Running some checks before starting writing
    """

    if not is_pandas and not is_polars:
        msg = "dataframe must be pandas or polars dataframe"
        raise PyreadstatError(msg)

    if variable_value_labels:
        for k,v in variable_value_labels.items():
            if type(v) != dict:
                msg = "variable_value_labels: value for key %s must be dict, got %s" % (k, str(type(v)))
                raise PyreadstatError(msg)

    cdef list valid_user_missing
    if missing_user_values:
        if file_format == FILE_FORMAT_DTA:
            valid_user_missing = valid_user_missing_stata
        elif file_format == FILE_FORMAT_SAS7BDAT or file_format == FILE_FORMAT_SAS7BCAT:
            valid_user_missing = valid_user_missing_sas

        for key, missing_values in missing_user_values.items():
            if not isinstance(missing_values, list):
                msg = "missing_user_values: values in dictionary must be list"
                raise PyreadstatError(msg)
            for val in missing_values:
                if val not in valid_user_missing:
                    msg = "missing_user_values supports values a to z for Stata and A to Z and _ for SAS, got %s instead" % str(val)
                    raise PyreadstatError(msg)

    if len(col_names) != len(set(col_names)):
        msg = "Non unique column names detected in the dataframe!"
        raise PyreadstatError(msg)

    for variable_name in col_names:
        if type(variable_name) != str:
                raise PyreadstatError("variable name '%s' is of type %s and it must be str (not starting with numbers!)" % (variable_name, str(type(variable_name))))
        if len(variable_name) == 0:
            raise PyreadstatError("variable names must be non-empty strings, not starting with numbers")
        if not variable_name[0].isalpha():
            raise PyreadstatError("variable name '%s' starts with an illegal (non-alphabetic) character: '%s' (ordinal %s)" % (variable_name, variable_name[0], ord(variable_name[0])))
        if " " in variable_name:
            raise PyreadstatError("variable name '%s' contains a space, which is not allowed" % variable_name)

    dirname = os.path.dirname(filename_bytes)
    if dirname and  not os.path.isdir(dirname):
        raise PyreadstatError(f"the destination folder {dirname} does not exist!")

cdef bytes filepath_to_bytes(object filename_path):
    """
    transforms an object with the path to the filename to bytes
    """
    if hasattr(os, 'fsencode'):
        try:
            filename_bytes = os.fsencode(filename_path)
        except UnicodeError:
            warnings.warn("file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly." % sys.getfilesystemencoding())
            filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")
    else:
        if type(filename_path) == str:
            filename_bytes = filename_path.encode('utf-8')
        elif type(filename_path) == bytes:
            filename_bytes = filename_path
        else:
            raise PyreadstatError("path must be either str or bytes")
    return filename_bytes


cdef int run_write(df, object filename_path, dst_file_format file_format, str file_label, object column_labels,
                   int file_format_version, object note, str table_name, dict variable_value_labels, 
                   dict missing_ranges, dict missing_user_values, dict variable_alignment,
                   dict variable_display_width, dict variable_measure, dict variable_format, bint row_compression) except *:
    """
    main entry point for writing all formats. Some parameters are specific for certain file type
    and are even incompatible between them. This function relies on the caller to select the right
    combination of parameters, not checking them otherwise.
    """

    cdef object natnamespace
    cdef object pd = None
    cdef bint is_pandas, is_polars
    cdef bytes filename_bytes
    cdef list col_names

    filename_bytes = filepath_to_bytes(filename_path)
    filename_bytes = os.path.expanduser(filename_bytes)

    df = nw.from_native(df, eager_only=True)
    natnamespace = nw.get_native_namespace(df)
    is_pandas = df.implementation.is_pandas()
    is_polars = df.implementation.is_polars()
    col_names = df.columns
    if is_pandas:
        pd = natnamespace

    initial_checks(is_pandas, is_polars, variable_value_labels, missing_user_values, file_format,
                         col_names, filename_bytes) 

    cdef readstat_error_t retcode
    cdef char *err_readstat
    cdef str err_message

    cdef readstat_writer_t *writer

    cdef bytes file_label_bytes
    cdef char *file_labl
    cdef int dta_str_max_len = 0


    if file_format == FILE_FORMAT_POR:
        col_names = [x.upper() for x in col_names]

    if file_format == FILE_FORMAT_DTA:
        if file_format_version >= 117:
            dta_str_max_len = dta_117_max_width
        elif file_format_version >= 111:
            dta_str_max_len = dta_111_max_width
        else:
            dta_str_max_len = dta_old_max_width

    cdef list col_types = get_narwhals_column_types(df, missing_user_values, variable_value_labels, dta_str_max_len)
    cdef int row_count = len(df)
    cdef int col_count = len(col_names)
    cdef dict col_names_to_types = {k:v[0] for k,v in zip(col_names, col_types)}

    cdef readstat_variable_t *variable
    cdef pywriter_variable_type curtype
    cdef int max_length
    cdef char *curformat
    cdef str tempformat
    cdef int col_indx
    cdef bytes cur_col_label
    cdef int col_label_count = 0

    cdef readstat_variable_t *tempvar
    cdef int row_indx
    cdef str curvalstr
    cdef double dtimelikeval
    #cdef np.ndarray values
    cdef object values
    cdef dict value_labels
    cdef int lblset_cnt = 0
    cdef readstat_label_set_t *label_set
    cdef list col_label_temp 
    cdef bint hasdatetime64
    cdef list pywriter_types, pywriter_timeunits
    cdef object df2
    cdef float mulfac, conv2secs
    cdef readstat_string_ref_t* strref
    cdef dict strref_map = dict()
    cdef int strref_cnt 
    cdef object strref_indx


    cdef int fd = open_file(filename_bytes)
    writer = readstat_writer_init()

    try:

        check_exit_status(readstat_set_data_writer(writer, write_bytes))

        if file_label:
            file_label_bytes = file_label.encode("utf-8")
            file_labl = <char *> file_label_bytes
            check_exit_status(readstat_writer_set_file_label(writer, file_labl))

        if note:
            if type(note) == str:
                note = [note]
            if type(note) == list:
                for line in note:
                    readstat_add_note(writer, line.encode("utf-8"))
            else:
                raise PyreadstatError(f"note should be either str or list, got {type(note)}")

        if file_format_version > -1:
            check_exit_status(readstat_writer_set_file_format_version(writer, file_format_version))

        if row_compression:
            check_exit_status(readstat_writer_set_compression(writer, READSTAT_COMPRESS_ROWS))

        # table name is used only for xpt files
        if table_name:
            table_name_bytes = table_name.encode("utf-8")
            tab_name = <char *> table_name_bytes
            check_exit_status(readstat_writer_set_table_name(writer, tab_name))

        # add variables
        if column_labels:
            if type(column_labels) != list and type(column_labels) != dict:
                raise PyreadstatError("column_labels must be either list or dict!")
            if type(column_labels) == dict:
                col_label_temp = list()
                for col_indx in range(col_count):
                    variable_name = col_names[col_indx]
                    if variable_name in column_labels.keys():
                        col_label_temp.append(column_labels[variable_name])
                    else:
                        col_label_temp.append(None)
                column_labels = col_label_temp

            col_label_count = len(column_labels)
            if col_label_count != col_count:
                raise PyreadstatError("length of column labels must be the same as number of columns")
     
        strref_cnt = 0
        for col_indx in range(col_count):
            curtype, max_length, _,_ = col_types[col_indx]
            variable_name = col_names[col_indx]
            # add variable
            variable = readstat_add_variable(writer, variable_name.encode("utf-8"), narwhals_to_readstat_types[curtype], max_length)
            # add format
            if variable_format:
                tempformat = variable_format.get(variable_name)
                if tempformat:
                   readstat_variable_set_format(variable, tempformat.encode("utf-8")) 
            if curtype in pyrwriter_datetimelike_types and (variable_format is None or variable_name not in variable_format.keys()):
                curformat = get_datetimelike_format_for_readstat(file_format, curtype)
                readstat_variable_set_format(variable, curformat)
            # prepare string_ref
            # for STRING_REF we have to add to a dict here before start writing
            if curtype == PYWRITER_DTA_STR_REF:
                for curval in df[variable_name]:
                    if curval not in strref_map:
                        curvalstr = str(curval)
                        strref = readstat_add_string_ref(writer, curvalstr.encode("utf-8"))
                        strref_map[curvalstr] = strref_cnt
                        strref_cnt += 1
            # labels
            if col_label_count:
                if column_labels[col_indx] is not None:
                    if type(column_labels[col_indx]) != str:
                        raise PyreadstatError("Column labels must be strings")
                    cur_col_label = column_labels[col_indx].encode("utf-8")
                    readstat_variable_set_label(variable, cur_col_label)
            if variable_value_labels:
                value_labels = variable_value_labels.get(variable_name)
                if value_labels:
                    labelset_name = variable_name + str(lblset_cnt)
                    lblset_cnt += 1
                    curuser_missing = None
                    if missing_user_values:
                        curuser_missing = missing_user_values.get(variable_name)
                    label_set = set_value_label(writer, value_labels, labelset_name,
                        col_names_to_types[variable_name], file_format, variable_name, curuser_missing)
                    readstat_variable_set_label_set(variable, label_set)
            # missing ranges
            if missing_ranges:
                cur_ranges = missing_ranges.get(variable_name)
                if cur_ranges:
                    if not isinstance(cur_ranges, list):
                        msg = "missing_ranges: values in dictionary must be list"
                        raise PyreadstatError(msg)
                    add_missing_ranges(cur_ranges, variable, curtype, variable_name)
            if variable_alignment:
                # At the moment this is ineffective for sav and dta (the function runs but in
                # the resulting file all alignments are still unknown)
                cur_alignment = variable_alignment.get(variable_name)
                if cur_alignment:
                    set_variable_alignment(variable, cur_alignment, variable_name)
            if variable_display_width:
                cur_display_width = variable_display_width.get(variable_name)
                if cur_display_width:
                    set_variable_display_width(variable, cur_display_width, variable_name)
            if variable_measure:
                cur_measure = variable_measure.get(variable_name)
                if cur_measure:
                    set_variable_measure(variable, cur_measure, variable_name)

        # start writing
        if file_format == FILE_FORMAT_SAS7BCAT:
            check_exit_status(readstat_begin_writing_sas7bcat(writer, &fd))
        elif file_format == FILE_FORMAT_DTA:
            check_exit_status(readstat_begin_writing_dta(writer, &fd, row_count))
        elif file_format == FILE_FORMAT_SAV:
            check_exit_status(readstat_begin_writing_sav(writer, &fd, row_count))
        elif file_format == FILE_FORMAT_POR:
            check_exit_status(readstat_begin_writing_por(writer, &fd, row_count))
        elif file_format == FILE_FORMAT_SAS7BDAT:
            check_exit_status(readstat_begin_writing_sas7bdat(writer, &fd, row_count))
        elif file_format == FILE_FORMAT_XPORT:
            check_exit_status(readstat_begin_writing_xport(writer, &fd, row_count))
        else:
            raise PyreadstatError("unknown file format")

        # validation
        check_exit_status(readstat_validate_metadata(writer))
        for col_indx in range(col_count):
            tempvar = readstat_get_variable(writer, col_indx)
            check_exit_status(readstat_validate_variable(writer, tempvar))

        # vectorized transform of datetime64ns columns
        pywriter_types = [x[0] for x in col_types]
        pywriter_timeunits = [x[3] for x in col_types]
        hasdatetime64 = PYWRITER_DATETIME64 in pywriter_types 
        hasdate64 = PYWRITER_DATE64 in pywriter_types 
        hastime64 = PYWRITER_TIME64 in pywriter_types 
        if hasdatetime64 or hasdate64 or hastime64:
            df2 = df.clone()
            if hasdatetime64:
                df2 = vectorized_convert_datetime_to_number(df2, file_format, pywriter_types, pywriter_timeunits, col_count)
            if hasdate64:
                df2 = vectorized_convert_date_to_number(df2, file_format, pywriter_types, col_count)
            if hastime64:
                df2 = vectorized_convert_time_to_number(df2, file_format, pywriter_types, col_count)
        else:
            df2 = df


        # inserting
        rowcnt = 0

        for row in df2.iter_rows():
            check_exit_status(readstat_begin_row(writer))

            for col_indx in range(col_count):

                tempvar = readstat_get_variable(writer, col_indx)
                curval = row[col_indx]
                curtype = pywriter_types[col_indx]
                is_missing = col_types[col_indx][2]
                curuser_missing = None
                if missing_user_values:
                    curuser_missing = missing_user_values.get(col_names[col_indx])
                
                if is_missing:
                    # For pandas we need isna because values can be Nan, NA, NAT, None
                    if is_pandas:
                        check_if_missing = pd.isna(curval)
                    # for other libraries the value would be None
                    else:
                        check_if_missing = curval is None
                    if check_if_missing:
                        check_exit_status(readstat_insert_missing_value(writer, tempvar))
                        continue

                if curuser_missing and curtype in pywriter_numeric_types:
                    if curval in curuser_missing:
                        check_exit_status(readstat_insert_tagged_missing_value(writer, tempvar, ord(curval)))
                        continue

                if curtype == PYWRITER_DOUBLE:
                    check_exit_status(readstat_insert_double_value(writer, tempvar, <double>curval))
                elif curtype == PYWRITER_INTEGER:
                    check_exit_status(readstat_insert_int32_value(writer, tempvar, curval))
                elif curtype == PYWRITER_LOGICAL:
                    check_exit_status(readstat_insert_int32_value(writer, tempvar, <int>curval))
                elif curtype == PYWRITER_CHARACTER:
                    check_exit_status(readstat_insert_string_value(writer, tempvar, curval.encode("utf-8")))
                elif curtype == PYWRITER_OBJECT:
                    curvalstr = str(curval)
                    check_exit_status(readstat_insert_string_value(writer, tempvar, curvalstr.encode("utf-8")))
                elif curtype == PYWRITER_DTA_STR_REF:
                    curvalstr = str(curval)
                    strref_indx = strref_map[curvalstr]
                    strref = readstat_get_string_ref(writer, strref_indx)
                    check_exit_status(readstat_insert_string_ref(writer, tempvar, strref))
                elif curtype == PYWRITER_DATETIME64 or curtype == PYWRITER_DATE64 or curtype == PYWRITER_TIME64:
                    check_exit_status(readstat_insert_double_value(writer, tempvar, <double>curval))
                elif curtype in pyrwriter_datetimelike_types:
                    dtimelikeval = convert_datetimelike_to_number(file_format, curtype, curval)
                    check_exit_status(readstat_insert_double_value(writer, tempvar, dtimelikeval))
                else:
                    raise PyreadstatError("Unknown data format to insert")

            check_exit_status(readstat_end_row(writer))
            rowcnt += 1

        check_exit_status(readstat_end_writing(writer))

    except:
        raise
    finally:
        readstat_writer_free(writer)
        close_file(fd)

    return 0



"""
Integration test: reading SAV file from HTTP server.
Uses Python's built-in http.server - no external dependencies.

Run with: pytest tests/test_http_integration.py -v
"""
import io
import os
import threading
import urllib.request
from contextlib import contextmanager
from http.server import HTTPServer, SimpleHTTPRequestHandler

import pyreadstat


@contextmanager
def http_server(directory):
    """Context manager that runs an HTTP server serving files from directory."""
    class QuietHandler(SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=directory, **kwargs)
        def log_message(self, *args):
            pass
    
    server = HTTPServer(("127.0.0.1", 0), QuietHandler)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    try:
        yield f"http://127.0.0.1:{server.server_address[1]}"
    finally:
        server.shutdown()


def test_read_sav_from_http():
    """Test reading SAV file from local HTTP server (simulates remote URL)."""
    data_folder = os.path.join(os.path.dirname(__file__), "..", "test_data", "basic")
    
    with http_server(data_folder) as base_url:
        url = f"{base_url}/sample.sav"
        with urllib.request.urlopen(url) as response:
            df, meta = pyreadstat.read_sav(io.BytesIO(response.read()))
        
        assert len(df) == 5, f"Expected 5 rows, got {len(df)}"
        assert len(df.columns) == 7, f"Expected 7 columns, got {len(df.columns)}"
